Require Export Base.

(** * Formulas and contexts *)

Definition var := nat.

Inductive form: Type :=
  | Var: var -> form
  | Fal: form
  | Imp: form -> form -> form
  | And: form -> form -> form
  | Or: form -> form -> form.

Definition Not s := Imp s Fal.
Definition Equi s t : form := And (Imp s t) (Imp t s).

Instance form_eq_dec : forall s t: form, dec (s = t).
Proof. unfold dec. repeat decide equality. Qed.

(** * Generic Entailment Relations *)

Section EntailmentRelationProperties.

  Variable F: Type.
  Variable E: list F -> F -> Prop.

  (*** Generic Structural Properties ***)
  Definition Monotonicity : Prop :=
    forall A A' s, A <<= A' -> E A s -> E A' s.

  Definition Reflexivity : Prop :=
    forall A s, s el A -> E A s.

  Definition Cut : Prop :=
    forall A s t, E A s -> E (s::A) t -> E A t.

  Definition Consistency : Prop :=
    exists s:F, ~E nil s.

End EntailmentRelationProperties.

Definition context := list form.

(*** Logical Properties ***)

Definition CharFal (E: context -> form -> Prop) : Prop :=
  forall A, E A Fal <-> forall s, E A s.

Definition CharImp (E: context -> form -> Prop) : Prop :=
  forall A s t, E A (Imp s t) <-> E (s::A) t.

Definition CharAnd (E: context -> form -> Prop) : Prop :=
  forall A s t, E A (And s t) <-> forall u, E (s::t::A) u -> E A u.

Definition CharOr (E: context -> form -> Prop) : Prop :=
  forall A s t, E A (Or s t) <-> forall u, E (s::A) u -> E (t::A) u -> E A u.

Fixpoint subst (sigma: var -> form) (s: form) : form :=
  match s with
  | Var x => sigma x
  | Fal => Fal
  | Imp s t => Imp (subst sigma s) (subst sigma t)
  | And s t => And (subst sigma s) (subst sigma t)
  | Or s t => Or (subst sigma s) (subst sigma t)
  end.

Definition Substitution (E: context -> form -> Prop) :=
  forall A s sigma, E A s -> E (map (subst sigma) A) (subst sigma s).

Definition EntailRelAllProps (E:  context -> form -> Prop) :=
  Reflexivity E /\ Monotonicity E /\ Cut E /\ Consistency E
  /\ CharImp E /\ CharFal E /\ CharAnd E /\ CharOr E
  /\ Substitution E.

Lemma EntailRelAllProps_ext E E':
  EntailRelAllProps E -> (forall A s, E A s <-> E' A s) -> EntailRelAllProps E'.
Proof.
  intros [R [M [C [Co [I [F [An [O S]]]]]]]] D. repeat split.
  - intros A s G. apply D. apply R, G. 
  - intros A A' s G H. apply D in H. apply D, (M A A' s G H).
  - intros A s t G H. apply D in G. apply D in H. apply D, (C A s t G H).
  - destruct Co as [s Co]. exists s. intros G. apply D in G. apply Co, G.
  - intros G. apply D. apply D in G. apply I, G.
  - intros G. apply D. apply D in G. apply I, G.
  - intros G s. apply D in G. apply D, F, G.
  - auto.
  - intros G u H. apply D in G. apply D in H. apply D. apply (An A s t); auto.
  - intros G. apply G. apply D. apply (An (s::t::A) s t).
    intros u H. apply C with (s:=t); auto. apply C with (s:=s); auto.
  - intros G u H K. apply D in G. apply D in H. apply D in K.
    apply D. apply (O A s t); auto.
  - intros G. apply G; apply D.
    + apply (O (s::A) s t). intros u H _. apply C with (s:=s); auto.
    + apply (O (t::A) s t). intros u _ H. apply C with (s:=t); auto.
  - intros A s sigma G. apply D in G. apply D. apply S, G.
Qed.

(** * Intuitionistic Gentzen System *)
Inductive gen: context -> form -> Prop :=
  | genV A x:       Var x el A -> gen A (Var x)
  | genF A u:       Fal el A -> gen A u
  | genIR A s t:    gen (s::A) t -> gen A (Imp s t)
  | genIL A s t u:  (Imp s t) el A -> gen A s -> gen (t::A) u -> gen A u
  | genAL A s t u:  (And s t) el A -> gen (s::t::A) u -> gen A u
  | genAR A s t:    gen A s -> gen A t -> gen A (And s t)
  | genOL A s t u:  (Or s t) el A -> gen (s::A) u -> gen (t::A) u -> gen A u
  | genOR1 A s t:   gen A s -> gen A (Or s t)
  | genOR2 A s t:   gen A t -> gen A (Or s t).

(* Hints for auto to use rules of gen *)
Hint Constructors gen.

(** Assumption rule *)
Lemma genA: Reflexivity gen.
Proof.
  intros A s. revert A. induction s; intros A C; auto.
  - apply genIR. eauto.
  - apply genAL with (s:=s1) (t:=s2); auto.
  - apply genOL with (s:=s1) (t:=s2); auto.
Qed.

Hint Resolve genA.

(** Weakening rule *)
Lemma gen_mono: Monotonicity gen.
Proof.
  intros A A' s C D. revert C. revert A'. induction D; intros A' C; eauto 9.
Qed.

Lemma genW A B s: gen A s -> A <<= B -> gen B s.
Proof. intros C D. apply gen_mono with (A:=A); auto. Qed.

(** Cut rule *)
Lemma gen_GCut A s B u:
  gen A s -> gen B u -> gen (A ++ rem B s) u.
Proof.
  intros C. revert A B u C.
  induction s as [x| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
  intros A B u C D.
  - remember (Var x) as v. induction C; subst; try inv Heqv; eauto using genW.
    + apply genIL with (s:=s) (t:=t); eauto using genW.
    + eauto 7.
  - remember Fal as v. induction C; subst; try congruence; eauto.
    + apply genIL with (s:=s) (t:=t); eauto using genW.
    + eauto 7.
  - remember (Imp s1 s2) as v. induction C; subst; try inv Heqv; eauto.
    + clear IHC. induction D as [B|B|B|B|B|B|B|B|B]; auto.
      * assert (Var x <> Imp s1 s2) by congruence. auto.
      * assert (Fal <> Imp s1 s2) by congruence. auto.
      * apply genIR. apply (genW IHD); auto.
      * { decide (Imp s t = Imp s1 s2) as [K|K].
          - inv K.
            assert (I: gen (A ++ rem B (Imp s1 s2)) s2) by (apply genW
              with (A:=(A ++ rem B (Imp s1 s2)) ++ rem (s1 :: A) s1); auto).
            apply genW with
            (A:=(A++rem B (Imp s1 s2)) ++ rem (A++rem (s2::B) (Imp s1 s2)) s2);
            auto. apply incl_app; auto. apply rem_app'; auto. 
            rewrite rem_comm. auto.
          - apply genIL with (s:=s) (t:=t); auto. apply (genW IHD2); auto. }
      * assert (And s t <> Imp s1 s2) by congruence.
        apply genAL with (s:=s) (t:=t); auto. apply (genW IHD); auto 9.
      * assert (Or s t <> Imp s1 s2) by congruence.
        apply genOL with (s:=s) (t:=t); auto.
          apply (genW IHD1); auto. apply (genW IHD2); auto.
    + apply genIL with (s:=s) (t:=t); eauto using genW.
    + eauto 7.
  - remember (And s1 s2) as v. induction C; subst; try congruence; eauto.
    + apply genIL with (s:=s) (t:=t); eauto using genW.
    + inv Heqv. clear IHC1 IHC2. induction D as [B|B|B|B|B|B|B|B|B]; auto.
      * assert (Var x <> And s1 s2) by congruence. auto.
      * assert (Fal <> And s1 s2) by congruence. auto.
      * apply genIR. apply (genW IHD); auto.
      * assert (Imp s t <> And s1 s2) by congruence.
        apply genIL with (s:=s) (t:=t); auto. apply (genW IHD2); auto.
      * { decide (And s t = And s1 s2) as [K|K].
          - inv K. remember (A ++ rem (s1 :: s2 :: B) (And s1 s2)) as Ax.
            assert (H2: gen (A ++ rem Ax s1) u) by auto.
            assert (H3: gen (A ++ rem (A ++ rem Ax s1) s2) u) by auto.
            apply (genW H3); auto. apply incl_app; auto. apply rem_app'; auto.
            subst Ax. unfold rem at 2. rewrite filter_app. apply rem_app'.
            + apply incl_appl. unfold rem.
              rewrite filter_and. apply filter_incl.
            + unfold rem. repeat rewrite filter_and.
              apply incl_appr. rewrite filter_fst'; try tauto.
              rewrite filter_fst'; try tauto. apply filter_pq_mono. tauto.
          - apply genAL with (s:=s) (t:=t); auto. apply (genW IHD); auto 9. }
      * assert (Or s t <> And s1 s2) by congruence.
        apply genOL with (s:=s) (t:=t); auto.
          apply (genW IHD1); auto. apply (genW IHD2); auto.
    + eauto 7.
  - remember (Or s1 s2) as v. induction C; subst; try congruence; eauto.
    + apply genIL with (s:=s) (t:=t); eauto using genW.
    + eauto 7.
    + inv Heqv. clear IHC. induction D as [B|B|B|B|B|B|B|B|B]; auto.
      * assert (Var x <> Or s1 s2) by congruence. auto.
      * assert (Fal <> Or s1 s2) by congruence. auto.
      * apply genIR. apply (genW IHD); auto.
      * assert (Imp s t <> Or s1 s2) by congruence.
        apply genIL with (s:=s) (t:=t); auto. apply (genW IHD2); auto.
      * assert (And s t <> Or s1 s2) by congruence.
        apply genAL with (s:=s) (t:=t); auto.
        apply genW with (A:= A ++ rem (s :: t :: B) (Or s1 s2)); auto 9.
      * { decide (Or s t = Or s1 s2) as [K|K].
          - inv K. apply genW with
                   (A:=A ++ rem (A ++ rem (s1 :: B) (Or s1 s2)) s1); auto 9.
            apply incl_app; auto. apply rem_app'; auto. rewrite rem_comm. auto.
          - apply genOL with (s:=s) (t:=t). auto.
            + apply (genW IHD1); auto. + apply (genW IHD2); auto. }
    + inv Heqv. clear IHC. induction D as [B|B|B|B|B|B|B|B|B]; auto.
      * assert (Var x <> Or s1 s2) by congruence. auto.
      * assert (Fal <> Or s1 s2) by congruence. auto.
      * apply genIR. apply (genW IHD); auto.
      * assert (Imp s t <> Or s1 s2) by congruence.
        apply genIL with (s:=s) (t:=t); auto. apply (genW IHD2); auto.
      * assert (And s t <> Or s1 s2) by congruence.
        apply genAL with (s:=s) (t:=t); auto.
        apply genW with (A:= A ++ rem (s :: t :: B) (Or s1 s2)); auto 9.
      * { decide (Or s t = Or s1 s2) as [K|K].
          - inv K. apply genW with
                   (A:=A ++ rem (A ++ rem (s2 :: B) (Or s1 s2)) s2); auto.
            apply incl_app; auto. apply rem_app'; auto. rewrite rem_comm. auto.
          - apply genOL with (s:=s) (t:=t). auto.
            + apply (genW IHD1); auto.
            + apply (genW IHD2); auto. }
Qed.

Lemma gen_cut: Cut gen.
Proof.
  intros A s t C D. apply genW with (A:=A ++ rem (s::A) s); auto.
  apply (gen_GCut C D).
Qed.

Lemma genCW A s t: gen A s -> gen [s] t -> gen A t.
Proof.
  intros D E. apply gen_cut with (s:=s); auto. apply (genW E). auto.
Qed.

Lemma gen_NoFal: ~ gen [] Fal.
Proof. intros C. inv C; auto. Qed.

Lemma gen_cons: Consistency gen.
Proof. exists Fal. apply gen_NoFal. Qed.

Lemma gen_imp: CharImp gen.
Proof. split.
  - intros C. apply gen_cut with (s:=Imp s t); eauto using genW.
  - auto.
Qed.

Lemma gen_fal: CharFal gen.
Proof. split.
  - intros C s. apply gen_cut with (s:=Fal); auto.
  - auto.
Qed.

Lemma gen_and: CharAnd gen.
Proof. split.
  - intros C u D. apply gen_cut with (s:=And s t). auto.
    apply genAL with (s:=s) (t:=t); auto. apply (genW D); auto 6.
  - auto.
Qed.

Lemma gen_and_both A s t: gen A (And s t) -> gen A s /\ gen A t.
Proof.
  intros C. destruct (gen_and A s t) as [H _]. pose (K:=H C).
  split; apply K; auto.
Qed.

Lemma gen_or: CharOr gen.
Proof. split.
  - intros C u D E. apply gen_cut with (s:=Or s t). auto.
    apply genOL with (s:=s) (t:=t). auto.
    + apply (genW D); auto.
    + apply (genW E); auto.
  - auto.
Qed.

Lemma gen_subst: Substitution gen.
Proof.
  intros A s sigma C. induction C; simpl; auto.
  - apply genA, in_map_iff. exists (Var x). auto.
  - apply gen_fal, genA, in_map_iff. exists Fal. auto.
  - apply genIL with (s:=subst sigma s) (t:=subst sigma t); auto.
    apply in_map_iff. exists (Imp s t). auto.
  - apply genAL with (s:=subst sigma s) (t:=subst sigma t); auto.
    + apply in_map_iff. exists (And s t). auto.
  - apply genOL with (s:=subst sigma s) (t:=subst sigma t); auto.
    + apply in_map_iff. exists (Or s t). auto.
Qed.

Lemma gen_EntailRelation: EntailRelAllProps gen.
Proof.
  split. apply genA.
  split. apply gen_mono.
  split. apply gen_cut.
  split. apply gen_cons.
  split. apply gen_imp.
  split. apply gen_fal.
  split. apply gen_and.
  split. apply gen_or.
  apply gen_subst.
Qed.

Lemma gen_NoVar x:
  ~ gen [] (Var x).
Proof. intros C. inv C; auto. Qed.

Lemma gen_NoNVar x:
  ~ gen [] (Not (Var x)).
Proof.
  intros C. apply gen_imp in C. inv C; auto; inv H; auto; congruence.
Qed.

(** * Intuitionistic Natural Deduction System *)

Inductive nd : context -> form -> Prop :=
  | ndA A s:      s el A -> nd A s
  | ndE A s:      nd A Fal -> nd A s
  | ndII A s t:   nd (s::A) t -> nd A (Imp s t)
  | ndIE A s t:   nd A (Imp s t) -> nd A s -> nd A t
  | ndAI A s t:   nd A s -> nd A t -> nd A (And s t)
  | ndAE A s t u: nd A (And s t) -> nd (s::t::A) u -> nd A u
  | ndOIL A s t:  nd A s -> nd A (Or s t)
  | ndOIR A s t:  nd A t -> nd A (Or s t)
  | ndOE A s t u: nd A (Or s t) -> nd (s::A) u -> nd (t::A) u -> nd A u.

(* Hints for auto to use rules of nd *)
Hint Constructors nd.

(* nd satisfies Entailment Relation properties *)
Lemma nd_refl: Reflexivity nd.
Proof. intros A s. auto. Qed.

Lemma nd_mono: Monotonicity nd.
Proof.
  intros A A' s C D. revert A' C. induction D; intros A' C; eauto.
  - eapply ndAE; try apply IHD2; auto.
  - eapply ndOE; auto.
Qed.

Lemma ndW A B s: nd A s -> A <<= B -> nd B s.
Proof. intros C D. apply nd_mono with (A:=A); auto. Qed.

Lemma nd_cut: Cut nd.
Proof. intros A s t C D. eauto. Qed.

Lemma ndCW A s t: nd A s -> nd [s] t -> nd A t.
Proof.
  intros D E. apply nd_cut with (s:=s); auto. apply (ndW E). auto.
Qed.

(* Lemma nd_cons: Consistency nd. --- will be proved using Gentzen *)

Lemma nd_imp: CharImp nd.
Proof. split; auto.
  intros C. apply ndIE with (s:=s); auto. apply (ndW C); auto.
Qed.

Lemma nd_fal: CharFal nd.
Proof. split; auto. Qed.

Lemma nd_and: CharAnd nd.
Proof. split; auto. intros C u. eauto. Qed.

Lemma nd_or: CharOr nd.
Proof. split; auto. intros C u. eauto. Qed.

Lemma nd_subst: Substitution nd.
Proof.
  intros A s sigma C. induction C; simpl; eauto.
  apply ndA, in_map_iff. exists s. auto.
Qed.

(** * Equivalence of Ni and Gi *)
Lemma gen_nd A s: gen A s -> nd A s.
Proof. intros C. induction C; eauto. Qed.

Lemma nd_gen A s: nd A s -> gen A s.
Proof.
  intros C. induction C; eauto.
  - apply gen_fal. auto.
  - apply gen_imp in IHC1. apply gen_cut with (s:=s); auto.
  - apply (gen_and A s t); auto.
  - apply (gen_or A s t); auto.
Qed.

Theorem gen_iff_nd A s: gen A s <-> nd A s.
Proof. split. apply gen_nd. apply nd_gen. Qed.

Lemma nd_NoFal: ~ nd [] Fal.
Proof. intros D. apply nd_gen in D. apply gen_NoFal, D. Qed.

Lemma nd_cons: Consistency nd.
Proof. exists Fal. apply nd_NoFal. Qed.

Lemma nd_EntailRelation: EntailRelAllProps nd.
Proof. apply (EntailRelAllProps_ext gen_EntailRelation gen_iff_nd). Qed.

(* One-sided ND *)
Fixpoint c2f A s: form :=
  match A with 
   | [] => s
   | s' :: A' => c2f A' (Imp s' s)
  end.

Lemma nd_onesided A s: nd A s <-> nd [] (c2f A s).
Proof.
  split; revert s; induction A as [|t A]; simpl; auto.
  intros s C. assert (D:= IHA (Imp t s) C). apply nd_imp, D.
Qed.

(** * Intuitionistic Hilbert System Hi *)

Definition FK (s t : form) : form :=
  Imp s (Imp t s).

Definition FS (s t u : form) : form :=
  (Imp (Imp s (Imp t u))
       (Imp (Imp s t)
            (Imp s u))).

Inductive hil (A : context): form -> Prop :=
  | hilK s t:     hil A (FK s t)
  | hilS s t u:   hil A (FS s t u)
  | hilAL s t:    hil A (Imp (And s t) s)
  | hilAR s t:    hil A (Imp (And s t) t)
  | hilAI s t:    hil A (Imp s (Imp t (And s t)))
  | hilOL s t:    hil A (Imp s (Or s t))
  | hilOR s t:    hil A (Imp t (Or s t))
  | hilOE s t u:  hil A (Imp (Imp s u) (Imp (Imp t u) (Imp (Or s t) u)))
  | hilE s:       hil A (Imp Fal s)
  | hilA s:       s el A -> hil A s
  | hilMP s t:    hil A (Imp s t) -> hil A s -> hil A t.

(** Equivalence of Ni and Hi *)
Lemma hil_nd A s :
  hil A s -> nd A s.

Proof.
  intros E. induction E as [s t|s t u|s t|s t|s t|s t|s t|
                            s t u|s|s E|s t E1 IHE1 E2 IHE2]; auto.
  - apply ndII, ndII, ndA. auto.
  - repeat apply ndII.
    apply ndIE with (s:= Imp t u).
    + apply ndII. apply ndIE with (s:=t); auto.
      apply ndIE with (s:=s); auto.
    + apply ndIE with (s:=s); auto.
  - apply ndII. apply ndAE with (s:=s) (t:=t); auto.
  - apply ndII. apply ndAE with (s:=s) (t:=t); auto. 
  - apply ndII, ndII. apply ndAI; auto.
  - repeat apply ndII. apply ndOE with (s:=s) (t:=t); auto.
    + apply ndIE with (s:=s); auto.
    + apply ndIE with (s:=t); auto.
  - apply ndIE with (s:=s) (t:=t); auto.
Qed.

Lemma hilAK A s t:
  hil A s -> hil A (Imp t s).
Proof. apply hilMP. apply hilK. Qed.

Lemma hilAS A s t u :
  hil A (Imp s (Imp t u)) -> hil A (Imp s t) -> hil A (Imp s u).
Proof. intros B. apply hilMP. revert B. apply hilMP. apply hilS. Qed.

Lemma hilI A s: hil A (Imp s s).
Proof.
  assert (E:= hilS A s (Imp s s) s).
  assert (F:= hilK A s (Imp s s)).
  assert (G:= hilK A s s).
  unfold FS, FK in *.
  exact (hilMP (hilMP E F) G).
Qed.

Lemma hilD w A v :
  hil (w::A) v -> hil A (Imp w v).
Proof.
  intros E. induction E as [s t|s t u|s t|s t|s t|s t|s t|
                            s t u|s|s E|s t E1 IHE1 E2 IHE2]; auto.
  - apply hilAK, hilK.
  - apply hilAK, hilS.
  - apply hilAK, hilAL.
  - apply hilAK, hilAR.
  - apply hilAK, hilAI.
  - apply hilAK, hilOL.
  - apply hilAK, hilOR.
  - apply hilAK, hilOE.
  - apply hilAK, hilE.
  - destruct E as [E|E].
    + subst. apply hilI.
    + apply hilAK. apply hilA, E.
  - apply hilAS with (t:=s); auto.
Qed.

Lemma nd_hil A s :
  nd A s -> hil A s.

Proof. 
  intros E. induction E as [A s E|A s E IH|A s t E IH|A s t E IHE F IHF|
                            A s t E IHE F IHF|A s t u E IHE F IHF|A s t E IH|
                            A s t E IH|A s t u E IHE F IHF G IHG]; auto.
  - apply hilA, E.
  - exact (hilMP (hilE A s) IH).
  - apply hilD, IH.
  - exact (hilMP IHE IHF).
  - exact (hilMP (hilMP (hilAI A s t) IHE) IHF).
  - apply hilMP with (s:=s).
    + apply hilMP with (s:=t).
      * apply hilD, hilD. auto.
      * exact (hilMP (hilAR A s t) IHE).
    + exact (hilMP (hilAL A s t) IHE).
  - exact (hilMP (hilOL A s t) IH).
  - exact (hilMP (hilOR A s t) IH).
  - apply hilMP with (s:=Or s t); auto.
    apply hilMP with (s:=Imp t u).
    + apply hilMP with (s:=Imp s u).
      * apply hilOE.
      * apply hilD. exact IHF.
    + apply hilD. exact IHG.
Qed.

Theorem hil_iff_nd A s:
  hil A s <-> nd A s.
Proof. split. apply hil_nd. apply nd_hil. Qed.


(** * Intuitionistic Tableaux (Fitting's System) *)

Inductive tab (C D: list form): Prop :=
 | tabF     : Fal el C     -> tab C D
 | tabC x   : Var x el C   -> Var x el D          -> tab C D
 | tabIP s t: Imp s t el C -> tab C (s::D)        -> tab (t::C) D   -> tab C D
 | tabIN s t: Imp s t el D -> tab (s :: C) [t]    -> tab C D
 | tabAP s t: And s t el C -> tab (s :: t :: C) D -> tab C D
 | tabAN s t: And s t el D -> tab C (s :: D)      -> tab C (t :: D) -> tab C D
 | tabOP s t: Or  s t el C -> tab (s :: C) D      -> tab (t :: C) D -> tab C D
 | tabON s t: Or  s t el D -> tab C (s :: t :: D) -> tab C D.

Lemma tabLW C D: tab C D -> forall C', C <<= C' -> tab C' D.
Proof. intros E. induction E; eauto 9 using tab. Qed.

Lemma tabRW C D: tab C D -> forall D', D <<= D' -> tab C D'.
Proof. intros E. induction E; eauto 9 using tab. Qed.

Lemma tabW C D C' D': tab C D -> C <<= C' -> D <<= D' -> tab C' D'.
Proof. intros E F G. apply tabLW with (C:=C); auto. apply (tabRW E), G. Qed.


Lemma tabCS C D s: s el C -> s el D -> tab C D.
Proof.
  revert C D. induction s as [x| |s IHs t IHt|s IHs t IHt|s IHs t IHt];
  intros C D E F.
  - apply (tabC E F).
  - apply (tabF D E).
  - apply (tabIP E).
    + apply tabIN with (s:=s) (t:=t); auto.
      apply tabIP with (s:=s) (t:=t); auto.
    + apply tabIN with (s:=s) (t:=t); auto.
  - apply (tabAP E); apply tabAN with (s:=s) (t:=t); auto.
  - apply (tabOP E); apply tabON with (s:=s) (t:=t); auto.
Qed.

Fixpoint OrAR (B: list form) : form :=
  match B with
  | [] => Fal
  | s :: B' => match B' with 
               | [] => s
               | _ => Or s (OrAR B')
               end
  end.

Lemma OrAR_mono A B: A <<= B -> gen [OrAR A] (OrAR B).
Proof.
  revert A. induction B as [|s B]; intros A D.
  - apply incl_nil_eq in D. subst. auto.
  - induction A as [|u A].
    + apply genF. auto.
    + apply incl_lcons in D as [D1 D2].
      specialize (IHA D2). destruct A as [|v A].
      * destruct D1 as [D1|D1]; subst; destruct B as [|v B]; simpl; auto.
        apply genOR2. apply (IHB [u]). auto.
      * apply genOL with (s:=u) (t:=OrAR (v::A)); auto.
        { destruct D1 as [D1|D1]; subst; destruct B as [|w B]; simpl; auto.
          apply genOR2. apply genW with (A:=[u]); auto. apply (IHB [u]). auto. }
        { apply (genW IHA); auto. }
Qed.


Lemma tab_genG C D:
  tab C D -> gen C (OrAR D).
Proof.
  intros E. induction E as [C D E|C D x E F|C D s t E _ IH1 _ IH2|
                            C D s t E _ IH|C D s t E _ IH|
                            C D s t E _ IH1 _ IH2|C D s t E _ IH1 _ IH2|
                            C D s t E _ IH]; eauto.
  - apply genW with (A:=[Var x]); auto.
    apply OrAR_mono with (A:=[Var x]). auto.
  - destruct D as [|w D].
    + apply genIL with (s:=s) (t:=t); auto.
    + apply (gen_or C s (OrAR (w :: D))); auto.
      apply genIL with (s:=s) (t:=t); auto. apply (genW IH2); auto.
  - simpl in IH. apply genCW with (s:= Imp s t).
    + apply genIR, IH.
    + apply OrAR_mono with (A:=[Imp s t]). auto.
  - destruct D as [|w D].
    + exfalso. apply E.
    + apply gen_cut with  (s:=(And (Or s (OrAR (w :: D)))
                          (Or t (OrAR (w :: D))))).
      * apply genAR; auto.
      * apply genAL with  (s:=Or s (OrAR (w :: D)))
                          (t:=Or t (OrAR (w :: D))); auto.
        apply genOL with  (s:=s) (t:=OrAR (w::D)); auto.
        apply genOL with  (s:=t) (t:=OrAR (w::D)); auto.
        apply genCW with  (s:=And s t); auto.
        apply OrAR_mono with (A:=[And s t]). auto.
  - destruct D as [|w D].
    + exfalso. apply E.
    + apply (gen_or C s (Or t (OrAR (w :: D)))); auto.
      * apply genCW with (s:=Or s t); auto.
        apply OrAR_mono with (A:=[Or s t]); auto.
      * apply genOL with (s:=t) (t:=(OrAR (w :: D))); auto.
        apply genCW with (s:=Or s t); auto.
        apply OrAR_mono with (A:=[Or s t]); auto.
Qed.

Lemma tab_gen A s:
  tab A [s] -> gen A s.
Proof. intros E. apply tab_genG in E. exact E. Qed.

Lemma gen_tab A s:
  gen A s -> tab A [s].
Proof.
  intros E. induction E as [A x E|A u E|A s t _ IH|A s t u E _ IH1 _ IH2|
                            A s t u E _ IH|A s t _ IH1 _ IH2|
                            A s t u E _ IH1 _ IH2|A s t _ IH|A s t _ IH];
                            eauto using tab, tabRW.
  apply tabAN with (s:=s) (t:=t); eauto using tabRW.
Qed.


Theorem gen_iff_tab A s:
  gen A s <-> tab A [s].
Proof.
  split. apply gen_tab. apply tab_gen.
Qed.

Lemma OrAR_mono_nd A B: A <<= B -> nd [OrAR A] (OrAR B).
Proof.
  revert A. induction B as [|s B]; intros A D.
  - apply incl_nil_eq in D. subst. auto.
  - induction A as [|u A].
    + simpl. auto.
    + apply incl_lcons in D as [D1 D2].
      specialize (IHA D2). destruct A as [|v A].
      * destruct D1 as [D1|D1]; subst; destruct B as [|v B]; simpl; auto.
        apply ndOIR. apply (IHB [u]). auto.
      * apply ndOE with (s:=u) (t:=OrAR (v::A)); auto.
        { destruct D1 as [D1|D1]; subst; destruct B as [|w B]; auto.
          - apply ndOIL. auto.
          - apply ndW with (A:=[u]); auto. apply ndOIR. apply (IHB [u]). auto. }
        { apply (ndW IHA); auto. }
Qed.

Lemma tab_ndG C D:
  tab C D -> nd C (OrAR D).
Proof.
  intros E. induction E as [C D E|C D x E F|C D s t E _ IH1 _ IH2|
                            C D s t E _ IH|C D s t E _ IH|
                            C D s t E _ IH1 _ IH2|C D s t E _ IH1 _ IH2|
                            C D s t E _ IH]; auto.
  - apply ndW with (A:=[Var x]); auto.
    apply OrAR_mono_nd with (A:=[Var x]). auto.
  - destruct D as [|w D].
    + apply ndIE with (s:=t) (t:=Fal).
      * apply ndII, (ndW IH2). auto.
      * apply ndIE with (s:=s) (t:=t); auto.
    + apply (nd_or C s (OrAR (w :: D))); auto. apply nd_cut with (s:=t).
      * apply ndIE with (s:=s) (t:=t); auto.
      * apply (ndW IH2); auto.
  - simpl in IH. apply ndCW with (s:= Imp s t).
    + apply ndII , IH.
    + apply OrAR_mono_nd with (A:=[Imp s t]). auto.
  - apply ndAE with (s:=s) (t:=t); auto.
  - destruct D as [|w D].
    + exfalso. apply E.
    + apply nd_cut with  (s:=(And (Or s (OrAR (w :: D)))
                          (Or t (OrAR (w :: D))))).
      * apply ndAI; auto.
      * apply ndAE with  (s:=Or s (OrAR (w :: D)))
                          (t:=Or t (OrAR (w :: D))); auto.
        apply ndOE with  (s:=s) (t:=OrAR (w::D)); auto.
        apply ndOE with  (s:=t) (t:=OrAR (w::D)); auto.
        apply ndCW with  (s:=And s t); auto.
        apply OrAR_mono_nd with (A:=[And s t]). auto.
  - apply ndOE with (s:=s) (t:=t); auto.
  - destruct D as [|w D].
    + exfalso. apply E.
    + apply (nd_or C s (Or t (OrAR (w :: D)))); auto.
      * apply ndCW with (s:=Or s t); auto.
        apply OrAR_mono_nd with (A:=[Or s t]); auto.
      * apply ndOE with (s:=t) (t:=(OrAR (w :: D))); auto.
        apply ndCW with (s:=Or s t); auto.
        apply OrAR_mono_nd with (A:=[Or s t]); auto.
Qed.

Lemma tab_nd A s:
  tab A [s] -> nd A s.
Proof. intros E. apply tab_ndG in E. exact E. Qed.


Lemma nd_OrCut A s t:
  nd A (Or s t) -> nd (s::A) t -> nd A t.
Proof. intros D E. apply ndOE with (s:=s) (t:=t); auto. Qed.

Lemma nd_OrARcut A B s:
  nd A (OrAR (s::B)) -> nd (s::A) (OrAR B) -> nd A (OrAR B).
Proof.
  destruct B as [|t B]; simpl; intros E F.
  - apply nd_cut with (s:=s); auto.
  - apply (nd_OrCut E F).
Qed.

(** * Decidability of Intuitionistic Gentzen *)

(** Subformula Closure *)

Definition sf_closed (A : context) : Prop :=
  forall u, u el A -> match u with
                       | Imp s t | And s t | Or s t => s el A /\ t el A
                       | _ => True
                      end.

Lemma sf_closed_app A B:
  sf_closed A -> sf_closed B -> sf_closed (A ++ B).
Proof.
  intros E F [| |s t|s t|s t] G; auto;
  apply in_app_or in G as [G|G];
  try apply E in G; try apply F in G; intuition.
Qed.

Lemma sf_closed_cons u s t A :
  (u = Imp s t \/ u = And s t \/ u = Or s t) ->
  s el A -> t el A -> sf_closed A -> sf_closed (u :: A).
Proof.
  intros [C|[C|C]]; subst; intros D E F [| |s' t'|s' t'|s' t'] G; auto;
  destruct G as [G|G]; try inv G; auto;
  apply F in G; destruct G; auto.
Qed.

(** Subformula-closed list *)

Fixpoint scl' (s : form) : context :=
  s :: match s with
         | Imp u v | And u v | Or u v => scl' u ++ scl' v
         | _ => nil
       end.

Lemma scl'_in s: s el scl' s.
Proof. destruct s; simpl; auto. Qed.

Lemma scl'_closed s: sf_closed (scl' s).
Proof.
  induction s; simpl; auto.
  - intros u [A|A]. inv A. auto. exfalso. auto.
  - intros u [A|A]. inv A. auto. exfalso. auto.
  - apply sf_closed_cons with (s:=s1) (t:=s2);
    auto using scl'_in, sf_closed_app.
  - apply sf_closed_cons with (s:=s1) (t:=s2);
    auto using scl'_in, sf_closed_app.
  - apply sf_closed_cons with (s:=s1) (t:=s2);
    auto using scl'_in, sf_closed_app.
Qed.

Fixpoint scl (A : context) : context :=
  match A with
    | nil => nil
    | s :: A' => scl' s ++ scl A'
  end.

Lemma scl_incl' A: A <<= scl A.
Proof. induction A as [|s A]; simpl; auto using scl'_in. Qed.

Lemma scl_incl A B: A <<= B -> A <<= scl B.
Proof. intros E. setoid_rewrite E. apply scl_incl'. Qed.

Lemma scl_closed A: sf_closed (scl A).
Proof.
  induction A as [|s A]; simpl. now auto.
  auto using scl'_closed, sf_closed_app.
Qed.


Definition goal := prod (context) form.

Instance goal_eq_dec (gamma delta: goal):
  dec (gamma = delta).
Proof. unfold dec. repeat decide equality. Qed.

Section Decidability.
  Variable A0 : context.
  Variable s0 : form.
  Definition U := scl (s0::A0).
  Definition U_sfc : sf_closed U := @scl_closed _.
  Definition Gamma := list_prod (power U) U.

  Definition step (Delta : list goal) (gamma : goal) : Prop :=
    let (A,u) := gamma in
    match u with
       | Var _ => u el A
       | Imp s t => (rep (s::A) U, t) el Delta
       | And s t => (A, s) el Delta /\ (A, t) el Delta
       | Or s t => (A, s) el Delta \/ (A, t) el Delta
       | _ => False
    end
    \/
    exists v, v el A /\
       match v with
         | Fal => True
         | Imp s t => (A, s) el Delta /\ (rep (t::A) U, u) el Delta
         | And s t => (rep (s::t::A) U, u) el Delta
         | Or s t => (rep (s::A) U, u) el Delta /\ (rep (t::A) U, u) el Delta
         | _ => False
        end.

  Instance step_dec Delta gamma :
    dec (step Delta gamma).
  Proof.
    destruct gamma as [A u]; simpl.
    apply or_dec.
    - destruct u as [x| |s t|s t|s t]; auto.
    - apply list_exists_dec. intros [x| |s t|s t|s t]; auto.
  Qed.

  Definition Lambda : list goal :=
    FCI.C (step := step) Gamma.

  Lemma lambda_closure gamma:
    gamma el Gamma -> step Lambda gamma -> gamma el Lambda.
  Proof. apply FCI.closure. Qed.

  Lemma lambda_ind (p : goal -> Prop) :
    (forall Delta gamma, inclp Delta p -> gamma el Gamma
                          -> step Delta gamma -> p gamma)
    -> inclp Lambda p.
  Proof. apply FCI.ind. Qed.

  Lemma gen_lambda A u:
    A <<= U -> u el U -> gen A u -> (rep A U,u) el Lambda.
  Proof.
    intros D1 D2 E.
    induction E as [A x F|A u F|A s t F IHF|A s t u F G IHG H IHH|
                    A s t u F G IHG|A s t F IHF G IHG|
                    A s t u F G IHG H IHH|A s t F IHF|A s t F IHF];
      (apply lambda_closure; [now eauto using in_prod, rep_power|]); simpl.
    - left. auto using rep_in.
    - right. exists Fal. auto using rep_in.
    - apply U_sfc in D2 as [D2 D3].
      left. rewrite rep_eq with (B := s::A); auto.
      setoid_rewrite rep_equi; auto.
    - assert (K := F). apply D1 in K.
      apply U_sfc in K as [K1 K2].
      right. exists (Imp s t); repeat split; auto using rep_in.
      rewrite rep_eq with (B := t::A); auto.
      setoid_rewrite rep_equi; auto.
    - assert (K := F). apply D1 in K.
      apply U_sfc in K as [K1 K2].
      right. exists (And s t); repeat split; auto using rep_in.
      rewrite rep_eq with (B := s::t::A); auto.
      setoid_rewrite rep_equi; auto.
    - apply U_sfc in D2 as [D2 D3]. left. auto.
    - assert (K := F). apply D1 in K.
      apply U_sfc in K as [K1 K2].
      right. exists (Or s t); repeat split; auto using rep_in.
      + rewrite rep_eq with (B := s::A); auto.
        setoid_rewrite rep_equi; auto.
      + rewrite rep_eq with (B := t::A); auto.
        setoid_rewrite rep_equi; auto.
    - apply U_sfc in D2 as [D2 D3]. left. auto.
    - apply U_sfc in D2 as [D2 D3]. left. auto.
  Qed.

  Lemma lambda_gen A u :
    (A,u) el Lambda -> gen A u.
  Proof.
    revert A u.
    cut (inclp Lambda (fun gamma => gen (fst gamma) (snd gamma))).
      { intros E A u. apply E. }
    apply lambda_ind. intros Delta [A u] E F G; simpl.
    assert (E': forall B v, (B,v) el Delta -> gen B v).
      { intros B v. apply E. } clear E.
    destruct G as [G|[[x| |s t|s t|s t] [G H]]].
    - destruct u; intuition; eauto using gen, genW, rep_incl.
    - contradiction H.
    - eauto.
    - intuition; eauto using genW, rep_incl.
    - intuition; eauto using genW, rep_incl.
    - assert (K1: gen (rep (s::A) U) u) by intuition.
      assert (K2: gen (rep (t::A) U) u) by intuition.
      eauto using genW, rep_incl.
  Qed.

  Theorem nd_dec: dec (nd A0 s0).
  Proof.
    apply (dec_prop_iff (gen_iff_nd _ _)).
    assert (A1: A0 <<= U).
    { apply scl_incl; auto. }
    assert (A2: s0 el U).
    { apply scl_incl'; auto. }
    apply dec_prop_iff with (X:= gen (rep A0 U) s0).
    - split; intros E; apply (genW E), rep_equi; auto.
    - decide ((rep A0 U, s0) el Lambda) as [E|E].
      + left. apply lambda_gen, E.
      + right. intros F. apply E.
        apply gen_lambda; auto.
        apply genW with (A := rep A0 U); auto.
        apply rep_equi, A1.
  Qed.

End Decidability.

(** Some underivablity results *)

Lemma gen_DN' A x s :
  A <<= [Var x; Not (Not (Var x))] -> gen A s -> s <> Fal /\ s <> Not (Var x).
Proof.
  intros H D.
  induction D as [A y E|A v E|A s t E IHE|A s t v E F IHF G _|
                  A s t u E|A s t|A s t u E|A s t|A s t].
  - split; discriminate.
  - exfalso. apply H in E. unfold In in E; intuition; discriminate.
  - split; intros J; inv J. cut (Fal <> Fal); auto. apply IHE. auto.
  - exfalso. apply H in E. destruct E as [E|[E|[]]]; inv E. apply IHF; auto.
  - exfalso. apply H in E. destruct E as [E|[E|[]]]; inv E.
  - unfold Not. split; congruence.
  - exfalso. apply H in E. destruct E as [E|[E|[]]]; inv E.
  - unfold Not. split; congruence.
  - unfold Not. split; congruence.
Qed.

Lemma gen_DN x :
  ~ gen [Not (Not (Var x))] (Var x).
Proof.
  intros E. inv E; unfold In in *; intuition; try discriminate.
  inv H2. apply gen_DN' with (x:=x) in H0; intuition.
Qed.

(** Formula size and depth *)

Fixpoint sizeF s : nat :=
  match s with
    | Imp s1 s2 | And s1 s2 | Or s1 s2 => 1 + sizeF s1 + sizeF s2
    | _ => 1
  end.

Fixpoint depth (s: form): nat :=
  match s with
  | Var _ => 0
  | Fal => 0
  | Imp s1 s2 | And s1 s2 | Or s1 s2 =>
      1 + (if decision (depth s1 <= depth s2) then depth s2 else depth s1)
  end.

(** Signed formulas and clauses *)

Inductive sform : Type :=
  | Pos : form -> sform
  | Neg : form -> sform.

Definition clause := list sform.

Notation "+ s" := (Pos s) (at level 35).
Notation "- s" := (Neg s).

Definition positive (S: sform): Prop :=
  match S with +s => True | -s => False end.

Definition negative (S: sform): Prop :=
  match S with +s => False | -s => True end.

Instance sform_eq_dec (S T: sform) : dec (S = T).
Proof. unfold dec. repeat decide equality. Defined.

Definition uns S : form :=
  match S with +s => s | -s => Not s end.

Fixpoint size C : nat :=
  match C with
    | nil => 0
    | +s::C' | -s::C' => sizeF s + size C'
  end.

Lemma uns_pos A :
  map uns (map Pos A) = A.
Proof.
  induction A; simpl; congruence.
Qed.

(** Incomplete formulas *)

Fixpoint FalFree u: Prop :=
  match u with
    | Var _ => True
    | Fal => False
    | Imp s t | And s t | Or s t => FalFree s /\ FalFree t
  end.

Fixpoint ImpFree u: Prop :=
  match u with
    | Var _ | Fal => True
    | Imp _ _ => False
    | And s t | Or s t => ImpFree s /\ ImpFree t
  end.

Fixpoint AndFree u: Prop :=
  match u with
    | Var _ | Fal => True
    | Imp s t => AndFree s /\ AndFree t
    | And _ _ => False
    | Or s t => AndFree s /\ AndFree t
  end.

Fixpoint OrFree u: Prop :=
  match u with
    | Var _ | Fal => True
    | Imp s t | And s t => OrFree s /\ OrFree t
    | Or _ _ => False
  end.

Section ssL.
  Definition ss_closed' (S: sform) (C: clause): Prop :=
      match S with
      | +Imp s t => -s el C /\ +t el C
      | -Imp s t => +s el C /\ -t el C
      | +And s t => +s el C /\ +t el C
      | -And s t => -s el C /\ -t el C
      | +Or  s t => +s el C /\ +t el C
      | -Or  s t => -s el C /\ -t el C
      | _ => True
      end.

  Definition ss_closed (C: clause) : Prop :=
    forall S, S el C -> ss_closed' S C.

  Definition s2b (S: sform): (bool * form) :=
    match S with | + s => (true, s) | - s => (false, s) end.

  Definition b2s (sign: bool) (s: form) := if sign then +s else -s.

  Fixpoint ssLb (sg: bool) (s: form): clause :=
    b2s sg s ::
    if sg then
      match s with
      | Imp s1 s2 => ssLb false s1 ++ ssLb true s2
      | And s1 s2 | Or  s1 s2 => ssLb true s1 ++ ssLb true s2
      | _ => []
      end
    else
      match s with
      | Imp s1 s2 => ssLb true s1 ++ ssLb false s2
      | And s1 s2 | Or  s1 s2 => ssLb false s1 ++ ssLb false s2
      | _ => []
      end.

  Definition ssL' (S: sform): clause := let (sg,s) := s2b S in ssLb sg s.

  Fixpoint ssL (C: clause): clause :=
    match C with
    | [] => []
    | Sf :: C' => ssL' Sf ++ ssL C'
    end.

  Lemma ss_closed'_mono S C C':
    ss_closed' S C -> C <<= C' -> ss_closed' S C'.
  Proof.
    intros E F. destruct S as [s|s]; destruct s; simpl in *; auto;
    destruct E as [E1 E2]; split; auto.
  Qed.

  Lemma ss_closed_app C C':
    ss_closed C -> ss_closed C' -> ss_closed (C ++ C').
  Proof.
    intros D E Sf F. apply in_app_iff in F as [F|F].
    - apply D in F. apply (ss_closed'_mono F). auto.
    - apply E in F. apply (ss_closed'_mono F). auto.
  Qed.

  Lemma ssL'_self (S: sform): ss_closed' S (ssL' S).
  Proof.
    destruct S as [s|s]; destruct s as [x| |s1 s2|s1 s2|s1 s2]; split;
    destruct s1, s2; simpl in *; auto.
  Qed.

  Lemma ssL'_in (S: sform): S el (ssL' S).
  Proof. destruct S as [s|s]; destruct s; simpl; auto. Qed.

  Lemma ssL'_correct (S: sform): ss_closed (ssL' S).
  Proof.
    intros Sg D. apply (@ss_closed'_mono Sg (ssL' Sg)).
    - apply ssL'_self.
    - revert Sg D.
      pose (Ssize (Sf: sform) := match Sf with | +s | -s => sizeF s end).
      pose (p S := forall Sg : sform, Sg el ssL' S -> ssL' Sg <<= ssL' S).
      apply (size_recursion Ssize (p:=p)).
      intros Sf D Sg E.
      destruct Sf as [s|s]; destruct s as [x| |s t|s t|s t];
      try (exfalso; apply E); destruct E as [E|E]; subst; auto;
      try (exfalso; apply E); unfold ssL' at 2; simpl;
      apply in_app_iff in E as [E|E].
      + assert (ssL' Sg <<= ssL' (-s)).
        { apply (D (-s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+t)).
        { apply (D (+t)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+s)).
        { apply (D (+s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+t)).
        { apply (D (+t)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+s)).
        { apply (D (+s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+t)).
        { apply (D (+t)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (+s)).
        { apply (D (+s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (-t)).
        { apply (D (-t)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (-s)).
        { apply (D (-s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (-t)).
        { apply (D (-t)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (-s)).
        { apply (D (-s)). simpl. omega. exact E. } auto.
      + assert (ssL' Sg <<= ssL' (-t)).
        { apply (D (-t)). simpl. omega. exact E. } auto.
  Qed.

  Lemma ssL_correct C: ss_closed (ssL C) /\ C <<= ssL C.
  Proof.
    induction C as [|Sf C [IH1 IH2]].
    - split; auto. intros Sf F. exfalso. apply F.
    - split; simpl.
      + apply ss_closed_app; auto using ssL'_correct.
      + assert (Sf el ssL' Sf) by (apply (ssL'_in Sf)). auto.
  Qed.

  Lemma ssL_in C S: S el C -> S el ssL C.
  Proof.
    intros E. induction C as [|S' C]; auto.
    destruct E as [E|E]; subst; simpl; apply in_app_iff.
    - left. apply ssL'_in.
    - right. auto.
  Qed.

  Lemma ssL'_in_ssL S C S': S' el ssL' S -> S el C -> S' el ssL C.
  Proof.
    revert S S'. induction C as [|Sx C]; intros S S' D E; auto.
    destruct E as [E|E]; subst; simpl; apply in_app_iff; auto.
    right. apply (IHC S S'); auto.
  Qed.

  Lemma ssL_mono C D: C <<= D -> ssL C <<= ssL D.
  Proof.
    revert D. induction C as [|S C]; intros D E.
    - simpl. auto.
    - destruct (ssL_correct D) as [D1 D2].
      apply incl_lcons in E as [E1 E2]. intros S' F.
      destruct S as [[x| |s t|s t|s t]|[x| |s t|s t|s t]];
      simpl in *; destruct F as [F|F]; subst; auto.
      + apply IHC; auto.
      + apply IHC; auto.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=+Imp s t); simpl; auto.
        * revert S' F. apply IHC, E2.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=+And s t); simpl; auto.
        * revert S' F. apply IHC, E2.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=+Or s t); simpl; auto.
        * revert S' F. apply IHC, E2.
      + revert S' F. apply IHC, E2.
      + revert S' F. apply IHC, E2.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=-Imp s t); simpl; auto.
        * revert S' F. apply IHC, E2.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=-And s t); simpl; auto.
        * revert S' F. apply IHC, E2.
      + apply in_app_iff in F as [F|F].
        * apply ssL'_in_ssL with (S:=-Or s t); simpl; auto.
        * revert S' F. apply IHC, E2.
  Qed.

  Lemma ssL_app C D: ssL (C ++ D) === ssL C ++ ssL D.
  Proof.
    induction C as [|S C]; simpl; auto.
    destruct S as [[x| |s t|s t|s t]|[x| |s t|s t|s t]]; simpl;
    rewrite IHC; auto; split; apply incl_shift; intros S E;
    apply in_app_iff in E as [E|E]; auto;
    apply in_app_iff in E as [E|E]; auto.
  Qed.

  Lemma ssL_idempotent C: ssL C === ssL (ssL C).
  Proof.
    split; intros S E.
    - destruct (ssL_correct (ssL C)). auto.
    - revert S E. induction C as [|Sx C]; intros S E; auto; simpl in *.
      apply ssL_app in E. apply in_app_iff in E as [E|E]; auto.
      apply in_app_iff. left.
      revert S E.
      pose (Ssize (Sf: sform) := match Sf with | +s | -s => sizeF s end).
      pose (p Sg := forall S : sform, S el ssL (ssL' Sg) -> S el ssL' Sg).
      apply (size_recursion Ssize (p:=p)).
      intros Sf D Sg E.
      destruct Sf as [[x| |s t|s t|s t]|[x| |s t|s t|s t]];
      simpl in *; destruct E as [E|E]; auto;
      apply in_app_iff in E as [E|E]; auto; right; rewrite ssL_app in E;
      apply in_app_iff in E as [E|E]; auto; apply in_app_iff.
      + left. apply (D (-s)); simpl; auto. omega.
      + right. apply (D (+t)); simpl; auto. omega.
      + left. apply (D (+s)); simpl; auto. omega.
      + right. apply (D (+t)); simpl; auto. omega.
      + left. apply (D (+s)); simpl; auto. omega.
      + right. apply (D (+t)); simpl; auto. omega.
      + left. apply (D (+s)); simpl; auto. omega.
      + right. apply (D (-t)); simpl; auto. omega.
      + left. apply (D (-s)); simpl; auto. omega.
      + right. apply (D (-t)); simpl; auto. omega.
      + left. apply (D (-s)); simpl; auto. omega.
      + right. apply (D (-t)); simpl; auto. omega.
  Qed.

  Lemma ssL'_in_ssL2 S C S': S' el ssL' S -> S el ssL C -> S' el ssL C.
  Proof.
    intros D E. assert (H:=ssL'_in_ssL D E). rewrite <- ssL_idempotent in H.
    exact H.
  Qed.

End ssL.
