Require Export PropiLmodel.

Lemma filter_idempotent
  X (p: X -> Prop) {p_dec: forall x, dec (p x)} (A: list X):
    filter p (filter p A) = filter p A.
Proof.
  apply filter_id. intros x E. apply in_filter_iff in E as [_ E]. exact E.
Qed.

Section ListOperation.

  Variable X: Type.
  Context {eq_decX: eq_dec X}.
  Implicit Type A B: list X.

  (* List union and intersection *)
  Fixpoint union A B :=
    match A with
    | [] => B
    | x :: A' => if decision (x el B) then union A' B
                                      else x :: union A' B
    end.

  Fixpoint inter A B :=
    match A with
    | [] => []
    | x :: A' => if decision (x el B) then x :: inter A' B
                                      else inter A' B
    end.

  Lemma union_sound A B (x:X):
    x el union A B <-> x el A \/ x el B.
  Proof. split.
    - induction A as [|y A]; simpl; auto.
      decide (y el B) as [_|_].
      + intros C. destruct (IHA C) as [I|I]; auto.
      + intros [C|C]; subst; auto. destruct (IHA C) as [I|I]; auto.
    - induction A as [|y A]; simpl; try tauto.
      intros [[C|C]|C]; decide (y el B) as [H|H]; subst; auto.
  Qed.

  Lemma inter_sound A B (x: X):
    x el inter A B <-> x el A /\ x el B.
  Proof. split.
    - induction A as [|y A]; simpl; try tauto.
      decide (y el B) as [H|_].
      + intros [C|C]; subst; auto. destruct (IHA C) as [C1 C2]; auto.
      + intros C. destruct (IHA C) as [C1 C2]; auto.
    - induction A as [|y A]; simpl; try tauto.
      decide (y el B) as [H|H]; intros [[C|C] D]; subst; auto.
      exfalso. auto.
  Qed.

  Fixpoint Binter (C: list (list X)): list X :=
    match C with
    | [] => []
    | A :: C' => inter A (Binter C')
    end.

  Lemma Binter_sound (C: list (list X)) (x: X):
    x el Binter C -> (forall A, A el C -> x el A).
  Proof.
    induction C as [|B C']; simpl.
    - intros F. exfalso. exact F.
    - intros E A [H|H].
      + subst. apply inter_sound in E as [E _]. exact E.
      + apply inter_sound in E as [_ E]. apply IHC'; auto.
  Qed.

  Fixpoint Bunion (C: list (list X)): list X :=
    match C with
    | [] => []
    | A :: C' => union A (Bunion C')
    end.

  Lemma Bunion_sound (C: list (list X)) (x: X):
    x el Bunion C <-> exists A, A el C /\ x el A.
  Proof.
    induction C as [|B C']; simpl; split.
    - intros F. exfalso. exact F.
    - intros [_ [F _]]. exact F.
    - intros E. apply union_sound in E as [E|E].
      + exists B. auto.
      + apply IHC' in E as [A [E1 E2]]. exists A. auto.
    - intros [A [[E|E] H]]; subst; apply union_sound.
      + left. exact H.
      + right. apply IHC'. exists A. auto.
  Qed.

  Lemma union_sub A B C: A <<= C /\ B <<= C <-> union A B <<= C.
  Proof. split.
    - intros [D E] x F. apply union_sound in F as [F|F]; auto.
    - intros E. split; intros x F; apply E; apply union_sound; auto.
  Qed.

  Lemma inter_sub A B C: A <<= C -> B <<= C -> inter A B <<= C.
  Proof. intros D E x F. apply inter_sound in F as [F1 F2]; auto. Qed.

  Lemma power_once A x: x el A -> [x] el power A.
  Proof.
    induction A as [|y A]; intros D.
    - exfalso. apply D.
    - destruct D as [D|D]; subst; simpl; auto.
      apply in_app_iff. right. apply in_map_iff. exists [].
      split; auto using power_nil.
  Qed.

End ListOperation.

Section UpwardClosure.

  Variable X: Type.
  Variable U: list X.
  Variable R: X -> X -> Prop.
  Context {eq_decX: eq_dec X}.
  Context {R_dec: forall x y, dec (R x y)}.
  Implicit Type x y: X.
  Implicit Type C: list X.

  (** Upward-closed sets *)
  Definition stepUC x C y : Prop := R x y.

  Definition UCS x: list X :=
    FCI.C (step := stepUC x) U.

  Lemma UCS_closure x y:
    y el U -> stepUC x (UCS x) y -> y el (UCS x).
  Proof. apply FCI.closure. Qed.

  Lemma UCS_ind x (p: X -> Prop):
    (forall C y, inclp C p -> y el U -> stepUC x C y -> p y)
    -> inclp (UCS x) p.
  Proof. apply FCI.ind. Qed.

  Lemma UCS_in x y:
    y el UCS x <-> y el U /\ R x y.
  Proof. split.
    - intros E. pattern y. apply (@UCS_ind x); auto.
    - intros [E F]. apply UCS_closure; auto.
  Qed.

  Lemma UCS_sub x: UCS x <<= U.
  Proof. intros y D. apply UCS_in in D as [D _]. exact D. Qed.

End UpwardClosure.

(** * Finitary Semantics *)

Definition upset (X: Type) (R: X -> X -> Prop) (A: list X) (K: list X):=
  A <<= K /\ forall x, x el A -> forall y, y el K -> R x y -> y el A.

Definition state := list var.

Definition base := list state.

Section TruthValues.

  Variable K: base.
  Variable Kdupfree: dupfree K.

  Definition truthVal (X: list state) := upset (fun X1 X2 => X1 <<= X2) X K.

  Lemma truthVal_union X Y: 
    truthVal X -> truthVal Y -> truthVal (union X Y).
  Proof.
    intros [E1 E2] [F1 F2]. split.
    - apply union_sub; auto.
    - intros sig D tau G H. apply union_sound. apply union_sound in D as [D|D].
      + left. apply (E2 sig); auto.
      + right. apply (F2 sig); auto.
  Qed.

  Lemma truthVal_inter X Y:
    truthVal X -> truthVal Y -> truthVal (inter X Y).
  Proof.
    intros [E1 E2] [F1 F2]. split.
    - apply (inter_sub E1 F1).
    - intros sig D tau G H. apply inter_sound in D as [D1 D2].
      apply inter_sound. split.
      + apply (E2 sig); auto.
      + apply (F2 sig); auto.
  Qed.

  Lemma truthVal_Bunion F:
    (forall X, X el F -> truthVal X) -> truthVal (Bunion F).
  Proof.
    intros E. split; intros sig D; apply Bunion_sound in D as [X [D1 D2]];
    destruct (E X D1) as [E1 E2]; auto.
    intros tau G H. apply Bunion_sound. exists X. split; auto.
    apply (E2 sig D2 tau G H).
  Qed.

  Lemma truthVal_Binter F:
    (forall X, X el F -> truthVal X) -> truthVal (Binter F).
  Proof.
    intros E. induction F as [|X F]; simpl; split; auto;
    assert (H: truthVal (Binter F)) by auto; destruct H as [H1 H2];
    intros sig D; apply inter_sound in D as [D1 D2]; auto.
    intros tau G I. apply inter_sound. split; auto.
    - assert (J: truthVal X) by auto. destruct J as [J1 J2].
      apply (J2 sig D1 tau G I).
    - apply (H2 sig D2 tau G I).
  Qed.

  (** Implication *)
  Section Implication.

    Definition stepImp (X Y: list state) (U: list (list state)) (Z: list state) 
      : Prop := inter Z X <<= Y.

    Definition TK: list (list state) := filter truthVal (power K).

    Definition FImpList X Y: list (list state) :=
      FCI.C (step := stepImp X Y) TK.

    Lemma FImpList_closure X Y Z:
      Z el TK -> stepImp X Y (FImpList X Y) Z -> Z el FImpList X Y.
    Proof. apply FCI.closure. Qed.

    Lemma FImpList_ind X Y (p: list state -> Prop):
      (forall C Z, inclp C p -> Z el TK -> stepImp X Y C Z -> p Z)
      -> inclp (FImpList X Y) p.
    Proof. apply FCI.ind. Qed.

    Lemma FImpList_in X Y Z:
      Z el FImpList X Y <-> Z el TK /\ inter Z X <<= Y.
    Proof. split.
      - intros E. pattern Z. apply (@FImpList_ind X Y); auto.
      - intros [E1 E2]. apply (@FImpList_closure X Y); auto.
    Qed.

    Definition FImp X Y := Bunion (FImpList X Y).

    Lemma truthVal_FImp X Y: truthVal (FImp X Y).
    Proof.
      split; intros sig D;
      unfold FImp in D; apply Bunion_sound in D as [Z [D1 D2]].
      - apply FImpList_in in D1 as [D1 _].
        unfold TK in D1. apply in_filter_iff in D1 as [_ [D1 D3]].
        apply D1. apply (D3 sig D2); auto.
      - intros tau G H. apply Bunion_sound. exists Z. split; auto.
        apply FImpList_in in D1 as [D1 _].
        unfold TK in D1. apply in_filter_iff in D1 as [_ [_ D1]].
        apply (D1 sig D2 tau G H).
    Qed.

    Lemma FImp_Heyting Z X Y: Z el TK ->
      (Z <<= FImp X Y <-> inter Z X <<= Y).
    Proof.
      intros D. split; intros E sig F.
      - apply inter_sound in F as [F1 F2].
        assert (H:= E sig F1). unfold FImp in H.
        apply Bunion_sound in H as [Z' [H1 H2]].
        apply FImpList_in in H1 as [_ H1].
        apply H1, inter_sound. auto.
      - apply Bunion_sound. exists Z. split; auto.
        apply FImpList_in. split; auto.
    Qed.

    Lemma FImp_greatest X Y Z:
      Z el TK -> inter Z X <<= Y -> Z <<= FImp X Y.
    Proof.
      intros D E sig G. apply Bunion_sound. exists Z. split; auto.
      apply FImpList_in. split; auto.
    Qed.

    Lemma FImp_in1 X Y A:
      A el FImp X Y -> A el X -> A el Y.
    Proof.
      intros D E. unfold FImp in D. apply Bunion_sound in D as [A' [D F]].
      apply FImpList_in in D as [_ D]. apply D.
      apply inter_sound. auto.
    Qed.

    Lemma FImp_in2 X Y A:
      truthVal X -> truthVal Y -> A el X -> A el Y -> A el FImp X Y.
    Proof.
      intros Dx Dy E F. apply Bunion_sound.
      assert (H: rep (inter (UCS K (R:=@incl var) A) X) K
                           === inter (UCS K (R:=@incl var) A) X).
        { apply rep_equi. intros B J. apply inter_sound in J as [J _].
          revert B J. apply UCS_sub. }
      exists (rep (inter (UCS K (R:=@incl var) A) X) K). split; auto.
      - apply FImpList_in. split.
        + apply in_filter_iff. split. apply rep_power. split.
          * unfold rep. apply filter_incl.
          * intros sig G tau H' I.
            rewrite H. rewrite H in G. apply inter_sound in G as [G1 G2].
            apply inter_sound. split.
              apply UCS_in. split; auto. apply UCS_in in G1 as [_ G1].
              intros x J. apply I, G1, J.
              destruct Dx as [_ D1]. apply (D1 sig G2 tau H' I).
        + intros sig G. apply inter_sound in G as [G1 _].
          rewrite H in G1. apply inter_sound in G1 as [G1 _].
          destruct Dy as [_ D1]. apply UCS_in in G1 as [G1 G3].
          apply (D1 A F sig G1 G3).
      - rewrite H. apply inter_sound. split; auto. apply UCS_in. split; auto.
        destruct Dx as [D1 _]. auto.
    Qed.

    Lemma FImp_in3 X Y A:
      truthVal Y -> A el Y -> A el FImp X Y.
    Proof.
      intros Dy F. apply Bunion_sound. exists (rep Y K). destruct Dy as [D1 D2].
      assert (rep Y K === Y) by (apply rep_equi, D1). split.
      - apply FImpList_in. split.
        + apply in_filter_iff. repeat split.
          * apply rep_power.
          * apply filter_incl.
          * intros x. rewrite H. intros I y. rewrite H. apply D2, I.
        + intros x I. apply inter_sound in I as [I _]. rewrite H in I. exact I.
      - rewrite H. exact F.
    Qed.

    Lemma FImp_in4 X Y A:
      (exists Z, Z el TK /\ A el Z /\ inter Z X <<= Y) -> A el FImp X Y.
    Proof.
      intros [Z [E1 [E2 E3]]]. apply (FImp_greatest E1) in E3. apply E3, E2.
    Qed.

  End Implication.

  Definition interpTK (x: var) : list state := filter (fun X => x el X) K.

  Fixpoint evalTK (s: form) : list state :=
    match s with
    | Var x => interpTK x
    | Fal => []
    | And s1 s2 => inter (evalTK s1) (evalTK s2)
    | Or  s1 s2 => union (evalTK s1) (evalTK s2)
    | Imp s1 s2 => FImp (evalTK s1) (evalTK s2)
    end.

  Fixpoint evalTK' (A: context) : list state :=
    match A with
    | [] => K
    | s :: A' => inter (evalTK' A') (evalTK s)
    end.

  Lemma truthVal_evalTK s: truthVal (evalTK s).
  Proof.
    induction s as [x| |s1 [IH1 IH2] s2 [IH3 IH4]|
                        s1 [IH1 IH2] s2 [IH3 IH4]|s1 [IH1 IH2] s2 [IH3 IH4]].
    - split; simpl.
      + unfold interpTK. apply filter_incl.
      + intros sig D tau E F. unfold interpTK in D.
        apply in_filter_iff in D as [D1 D2]. apply in_filter_iff. auto.
    - simpl. split; auto.
    - split; simpl; intros sig D;
      unfold FImp in D; apply Bunion_sound in D as [Z [D1 D2]].
      + apply FImpList_in in D1 as [D1 _].
        unfold TK in D1. apply in_filter_iff in D1 as [D1 _].
        apply power_incl in D1. apply D1, D2.
      + intros tau E F. apply Bunion_sound. exists Z. split; auto.
        apply FImpList_in in D1 as [D1 _].
        unfold TK in D1. apply in_filter_iff in D1 as [_ [_ D1]].
        apply (D1 sig D2 tau E F).
    - split; simpl; intros sig D;
      apply inter_sound in D as [D1 D2].
      + apply IH1, (IH2 sig); auto.
      + intros tau E F. apply inter_sound. split.
        * apply (IH2 sig D1 tau E F).
        * apply (IH4 sig D2 tau E F).
    - split; simpl; intros sig D.
      + apply union_sound in D as [D|D].
        * apply IH1, (IH2 sig D); auto.
        * apply IH3, (IH4 sig D); auto.
      + intros tau E F. apply union_sound. apply union_sound in D as [D|D].
        * left. apply (IH2 sig D tau E F).
        * right. apply (IH4 sig D tau E F).
  Qed.

  Lemma truthVal_evalTK' A: truthVal (evalTK' A).
  Proof.
    induction A as [|s A [IH1 IH2]]; simpl; split; auto; intros sig D;
    apply inter_sound in D as [D1 D2]; auto.
    intros tau E F. apply inter_sound. split.
    - apply (IH2 sig D1 tau E F).
    - destruct (truthVal_evalTK s) as [_ H]. apply (H sig D2 tau E F).
  Qed.

  Lemma rep_evalTK s: evalTK s === rep (evalTK s) K.
  Proof.
    destruct (truthVal_evalTK s) as [H _]. rewrite (rep_equi H). auto.
  Qed.

  Lemma rep_evalTK' A: evalTK' A === rep (evalTK' A) K.
  Proof.
    destruct (truthVal_evalTK' A) as [H _]. rewrite (rep_equi H). auto.
  Qed.

  Lemma nd_soundTK A s: nd A s -> evalTK' A <<= evalTK s.
  Proof.
    intros C. induction C as [A s D|A s _ IH|A s t _ IH|A s t _ IH1 _ IH2|
                              A s t _ IH1 _ IH2|A s t u _ IH1 _ IH2|A s t _ IH|
                              A s t _ IH|A s t u _ IH1 _ IH2 _ IH3]; simpl in *.
    - induction A as [|t A].
      + exfalso. apply D.
      + simpl. intros sig E. apply inter_sound in E as [E1 E2].
        destruct D as [D|D]; subst; auto. apply (IHA D sig E1).
    - apply incl_nil_eq in IH. rewrite IH. auto.
    - assert (H1:= rep_evalTK' A). destruct (truthVal_evalTK' A) as [H2 H3].
      intros sig E. apply Bunion_sound.
      exists (rep (evalTK' A) K). split.
      + apply FImpList_in. split.
        * unfold TK. apply in_filter_iff. repeat split.
          { apply rep_power. } { rewrite <- H1. auto. }
          { intros tau1. rewrite <- H1. intros F tau2 G I.
            rewrite <- H1. apply (H3 tau1 F tau2 G I). }
        * intros tau F. apply inter_sound in F as [F1 F2]. rewrite <- H1 in F1.
          apply IH, inter_sound. auto.
      + rewrite <- H1. exact E.
    - assert (H1:= rep_evalTK' A). rewrite H1 in IH1.
      apply FImp_Heyting in IH1.
      + intros sig E. apply IH1. apply inter_sound. rewrite <- H1. auto.
      + apply in_filter_iff. repeat split.
        * apply rep_power.
        * apply filter_incl.
        * intros sig. rewrite <- H1. intros F tau G I. rewrite <- H1.
          destruct (truthVal_evalTK' A) as [_ H2].
          apply (H2 sig F tau G I).
    - intros sig E. apply inter_sound. auto.
    - rewrite <- IH2. intros sig E. specialize (IH1 sig E).
      apply inter_sound in IH1 as [H1 H2].
      apply inter_sound. split; auto. apply inter_sound. auto.
    - intros sig E. apply union_sound. auto.
    - intros sig E. apply union_sound. auto.
    - intros sig E. specialize (IH1 sig E). apply union_sound in IH1 as [H|H].
      + apply IH2. apply inter_sound. auto.
      + apply IH3. apply inter_sound. auto.
  Qed.

End TruthValues.

Section CounterModels.

  Let K1: base := [[]].

  Example K_Fal: ~ nd [] Fal.
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K1) E).
    apply incl_nil_eq in H. discriminate H.
  Qed.

  Example K_var x: ~ nd [] (Var x).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K1) E).
    apply incl_nil_eq in H. discriminate H.
  Qed.
 
  Example K_orxy: ~ nd [] (Or (Var 0) (Var 1)).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K1) E).
    apply incl_nil_eq in H. discriminate H.
  Qed.

  Let K2: base := [[0]].

  Example K_neg: ~ nd [] (Not (Var 0)).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K2) E).
    apply (H [0]). simpl. auto.
  Qed.

  Let K3: base := [[]; [0]].

  Example K_XM: ~ nd [] (Or (Var 0) (Not (Var 0))).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K3) E). cbv in H.
    assert (K:= H []). assert (I: [0] = [] \/ False) by auto.
    destruct I as [I|I]; auto; discriminate I.
  Qed.

  Example K_DN: ~ nd [] (Imp (Not (Not (Var 0))) (Var 0)).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K3) E). cbv in H.
    assert (K:= H []). assert (I: [0] = [] \/ False) by auto.
    destruct I as [I|I]; auto; discriminate I.
  Qed.

  Let K4: base := [[]; [0]; [0;1]].

  Example K_Peirce: ~ nd [] (Imp (Imp (Imp (Var 0) (Var 1)) (Var 0)) (Var 0)).
  Proof.
    intros E. assert (H:= nd_soundTK (K:=K4) E). cbv in H.
    assert (K:= H []). assert (I: [0] = [] \/ [0; 1] = [] \/ False) by auto.
    destruct I as [I|[I|I]]; auto; discriminate I.
  Qed.

End CounterModels.


Section Demo.

  Definition Hintikka (A B: context): Prop := 
    (forall s, s el A ->
      match s with
      | Fal => False
      | Imp s1 s2 => s1 el B \/ s2 el A
      | And s1 s2 => s1 el A /\ s2 el A
      | Or  s1 s2 => s1 el A \/ s2 el A
      | Var x => ~ Var x el B
      end) /\
    (forall s, s el B ->
      match s with
      | And s1 s2 => s1 el B \/ s2 el B
      | Or  s1 s2 => s1 el B /\ s2 el B
      | _ => True
      end).

  Lemma Hintikka_equi (C D C' D': context):
    Hintikka C D -> C' === C -> D' === D -> Hintikka C' D'.
  Proof.
    intros [E1 E2] F G. split; intros s.
    - rewrite F. intros H. specialize (E1 s H).
      destruct s; auto; try rewrite F; auto; rewrite G; auto.
    - rewrite G. intros H. specialize (E2 s H).
      destruct s; auto; rewrite G; auto.
  Qed.

  Instance Hintikka_dec C D: dec (Hintikka C D).
  Proof. apply and_dec; apply list_forall_dec; intros [| | | |]; auto. Defined.

  Definition PContext: Type := prod context context.

  Definition Demo (D: list PContext): Prop :=
    forall A B, (A, B) el D ->
      Hintikka A B /\ 
      (forall s t, Imp s t el B ->
        exists C', C' el D /\ let (A', B') := C' in
                                A <<= A' /\ s el A' /\ t el B').

  Definition dsub (D D': list PContext): Prop :=
    forall A B, (A, B) el D -> exists C', C' el D' /\
      let (A', B'):= C' in A === A' /\ B === B'.

  Definition dequi D D' : Prop := dsub D D' /\ dsub D' D.

  Lemma Demo_equi D D': Demo D -> dequi D D' -> Demo D'.
  Proof.
    intros E [F1 F2] A B G.
    destruct (F2 A B G) as [[A' B'] [H1 [H2 H3]]].
    destruct (E A' B' H1) as [H4 H5]. split.
    - apply (Hintikka_equi H4 H2 H3).
    - intros s t. rewrite H3. intros K.
      destruct (H5 s t K) as [[A1 B1] [K1 [K2 [K3 K4]]]].
      destruct (F1 A1 B1 K1) as [[A2 B2] [I1 [I2 I3]]].
      exists (A2,B2). repeat split; auto.
      + rewrite H2. rewrite <- I2. exact K2.
      + rewrite <- I2. exact K3.
      + rewrite <- I3. exact K4.
  Qed.

End Demo.


Instance prod_eq_dec X Y:
  eq_dec X -> eq_dec Y -> eq_dec (prod X Y).
Proof.
  intros D E. intros [x1 y1] [x2 y2].
  decide (x1 = x2) as [H|H]; decide (y1 = y2) as [K|K];
  try (left; subst; reflexivity); right; intros I; inv I; auto.
Defined.


Section RelationToList.

  Variable X: Type.
  Variable R: X -> X -> Prop.
  Variable S: list X.
  Context {eq_decX: eq_dec X}.
  Context {R_dec: forall x y, dec (R x y)}.

  Let SS := list_prod S S.

  Definition stepRList (C: list (X * X)) (p: X * X): Prop :=
    let (x,y) := p in x el S /\ y el S /\ R x y.

  Instance stepRList_dec C p: dec (stepRList C p).
  Proof. unfold stepRList. destruct p as [x y]. auto. Defined.

  Definition RList : list (X * X) := FCI.C (step:= stepRList) SS.

  Lemma RList_closure p:
    p el SS -> stepRList RList p -> p el RList.
  Proof. apply FCI.closure. Qed.

  Lemma RList_ind (P: X * X -> Prop):
    (forall C p, inclp C P -> p el SS -> stepRList C p -> P p)
    -> inclp RList P.
  Proof. apply FCI.ind. Qed.

  Lemma RList_in p: p el RList <-> let (x,y) := p in x el S /\ y el S /\ R x y.
  Proof. split.
    - intros E. pattern p. apply RList_ind; auto.
    - intros E. apply RList_closure; auto. destruct p as [x y].
      destruct E as [E1 [E2 _]]. apply in_prod_iff. auto.
  Qed.

End RelationToList.

Section FiniteClosure.

  Variable F: Type.
  Context {eq_decF: eq_dec F}.
  Variable FS: list F.
  Variable FE: list (F * F).
  Implicit Type x y z: F.

  (** Reflexive-transitive closure *)
  Definition stepRTC (C: list (F * F)) (p: F * F) : Prop :=
    p el FE \/
    let (x, y) := p in
    y = x \/
    (exists z, z el FS /\ (x,z) el C /\ (z,y) el C).

  Instance stepRTC_dec C p:
    dec (stepRTC C p).
  Proof. destruct p as [x y]. apply or_dec; auto. Defined.

  Definition RTC: list (F * F) :=
    FCI.C (step := stepRTC) (list_prod FS FS).

  Lemma RTC_closure p:
    p el list_prod FS FS -> stepRTC RTC p -> p el RTC.
  Proof. apply FCI.closure. Qed.

  Lemma RTC_ind (P: (F*F) -> Prop):
    (forall C p, inclp C P -> p el list_prod FS FS -> stepRTC C p -> P p)
    -> inclp RTC P.
  Proof. apply FCI.ind. Qed.

  Lemma RTC_reflexive x: x el FS -> (x,x) el RTC.
  Proof.
    intros E. apply RTC_closure.
    - apply in_prod_iff. auto.
    - right. left. auto.
  Qed.

  Lemma RTC_transitive x y z:
    x el FS -> y el FS -> z el FS
    -> (x,y) el RTC -> (y,z) el RTC -> (x,z) el RTC.
  Proof.
    intros Ex Ey Ez G H. apply RTC_closure.
    - apply in_prod_iff. auto.
    - right. right. exists y. auto.
  Qed.

  Lemma RTC_in x y: (x,y) el RTC -> x el FS /\ y el FS.
  Proof.
    intros E.
    pose (P (p: F*F) := let (u,v):= p in u el FS /\ v el FS).
    change (P (x,y)). apply RTC_ind; auto.
    intros C [u v] D G _. apply in_prod_iff in G. exact G.
  Qed.

  Lemma RTC_iff x y:
    (x,y) el RTC <-> x el FS /\ y el FS /\
                      ((x,y) el FE \/ y = x \/
                       (exists z, z el FS /\ (x,z) el RTC /\ (z,y) el RTC)).
  Proof. split.
    - intros E. assert (H:= RTC_in E). destruct H as [H1 H2].
      pose (P (p: F*F) := let (a,b) := p in
              a el FS /\ b el FS /\
              ((a, b) el FE \/ b = a \/
              (exists z : F,
                 z el FS /\ (a, z) el RTC /\ (z, b) el RTC))).
      change (P (x,y)). apply RTC_ind; auto.
      intros C (x', y') G H K. apply in_prod_iff in H as [H3 H4].
      destruct K as [K|[K|[z [K1 [K2 K3]]]]]; repeat split; auto.
      right. right. exists z. repeat split; auto;
      apply RTC_closure; try (apply in_prod_iff; auto).
      + destruct (G (x', z) K2) as [_ [_ [Gxz|[Gxz|Gxz]]]];
        unfold stepRTC; auto.
      + destruct (G (z, y') K3) as [_ [_ [Gzy|[Gzy|Gzy]]]];
        unfold stepRTC; auto.
    - intros [D [E G]]. apply RTC_closure; auto.
      apply in_prod_iff; auto.
  Qed.

  (** Compute UCS from RTC *)
  Definition ucsPred x (p: F * F) := let (x',_) := p in x' = x.

  Lemma dec_ucsPred x : forall p, dec (ucsPred x p).
  Proof. intros [x' y]. auto. Defined.

  Let UCSFilter x FERTC : list (F * F)
     := filter (ucsPred x) (p_dec:=dec_ucsPred x) FERTC.

  Definition UCSE x FERTC:=
    map (fun (p: F * F) => let (_,y) := p in y) (UCSFilter x FERTC).

  Lemma UCSE_sound x y: (x,y) el RTC <-> y el UCSE x RTC.
  Proof.
    unfold UCSE. split; intros E.
    - apply in_map_iff. exists (x,y). split; auto.
      apply in_filter_iff. unfold ucsPred. split; auto.
    - apply in_map_iff in E as [[x' y'] [G H]]. subst.
      unfold UCSFilter in H. apply in_filter_iff in H as [H1 H2].
      unfold ucsPred in H2. subst. exact H1.
  Qed.

  (* By stepFKImp we assume FE is FERTC, i.e. FE satisfies RTC *)
  Definition stepFKImp (A B C: list F) x : Prop :=
    forall y, y el inter (UCSE x FE) A -> y el B.

  Definition FKImp A B: list F :=
    FCI.C (step := stepFKImp A B) FS.

  Lemma FKImp_closure A B x:
    x el FS -> stepFKImp A B (FKImp A B) x -> x el FKImp A B.
  Proof. apply FCI.closure. Qed.

  Lemma FKImp_ind A B (p: F -> Prop):
    (forall C x, inclp C p -> x el FS -> stepFKImp A B C x -> p x)
    -> inclp (FKImp A B) p.
  Proof. apply FCI.ind. Qed.

  Lemma FKImp_in x A B:
    x el FKImp A B
    <-> x el FS /\ forall y, y el inter (UCSE x FE) A -> y el B.
  Proof. split.
    - intros E. pattern x. apply (@FKImp_ind A B); auto.
    - intros [E1 E2]. apply (@FKImp_closure A B); auto.
  Qed.

End FiniteClosure.

Section DemoKBase.

  Fixpoint CVars (A: context): list var :=
    match A with
    | [] => []
    | s :: A' => match s with
                 | Var x => x :: CVars A'
                 | _ => CVars A'
                 end
    end.

  Lemma CVars_correct A x: (Var x) el A <-> x el CVars A.
  Proof.
    induction A as [|s A]; try tauto.
    destruct s as [| | | |]; simpl; split; try tauto;
    intros [H|H]; try tauto; try inv H; auto.
  Qed.

  Definition baseVars (D: list (context * context)): list var :=
    undup (Bunion
            (map (fun (C: context * context) => let (A,_):= C in CVars A) D)).

  Lemma baseVars_incl (D: list (context * context)) C:
    C el D -> let (A,_):= C in CVars A <<= baseVars D.
  Proof.
    destruct C as [A B]. intros E. unfold baseVars. rewrite undup_id_equi.
    intros x F. apply Bunion_sound. exists (CVars A). split; auto.
    apply in_map_iff. exists (A, B). auto.
  Qed.

  Lemma baseVars_in D A B:
    (A, B) el D -> rep (CVars A) (baseVars D) === CVars A.
  Proof. intros E. rewrite rep_equi; auto. apply (baseVars_incl E). Qed.

  Definition baseKD (D: list (context * context)): list state :=
    undup 
      (map (fun (C: context * context) =>
              let (A,B) := C in rep (CVars A) (baseVars D)) D).

  Lemma baseKD_in D: forall A B,
    (A, B) el D -> rep (CVars A) (baseVars D) el baseKD D.
  Proof.
    intros A B E. unfold baseKD. rewrite undup_id_equi.
    apply in_map_iff. exists (A, B). split; auto.
  Qed.

  Lemma baseKD_sound D (FA: list var):
    FA el baseKD D -> exists C, C el D /\ let (A, B) := C in CVars A === FA.
  Proof.
    intros E.
    unfold baseKD in E. rewrite undup_id_equi in E.
    apply in_map_iff in E as [[A' B'] [F G]]. exists (A', B').
    split; auto. rewrite <- F. symmetry.
    apply rep_equi. apply (baseVars_incl G).
  Qed.

  Lemma baseKD_complete D A B:
    (A, B) el D -> rep (CVars A) (baseVars D) el baseKD D
                  /\ CVars A === rep (CVars A) (baseVars D).
  Proof.
    intros E. split.
    - unfold baseKD. rewrite undup_id_equi. apply in_map_iff.
      exists (A,B). auto.
    - symmetry. apply rep_equi. apply (baseVars_incl E).
  Qed.

  Definition satisD D A s: Prop :=
    rep (CVars A) (baseVars D) el (evalTK (baseKD D) s).

End DemoKBase.

Section PContextRelation.

  Definition RC (C C': PContext): Prop
    := let (A,_) := C in let (A', _) := C' in A <<= A'.

  Instance RC_dec: forall C C', dec (RC C C').
  Proof. unfold RC. intros [A _] [A' _]. auto. Defined.

  Definition RCList (D: list PContext) := (RList (R:=RC) D).

  Let pRC (x: var) (C: PContext) := let (A,_):= C in Var x el A.

  Let pRC_dec x: forall C, dec (pRC x C). Proof. intros [A B]. auto. Defined.

  Definition RCinterp (D: list PContext) x
    : list (context * context) := filter (p_dec:= pRC_dec x) (pRC x) D.

  Lemma RTC_RCList (D: list PContext) (C1 C2: PContext):
    (C1, C2) el @RTC _ _ D (RCList D) <-> C1 el D /\ C2 el D /\ RC C1 C2.
  Proof. split.
    - intros E. destruct (RTC_in E) as [E1 E2]. repeat split; auto.
      pose (p (P: PContext * PContext) := let (C,C') := P in RC C C').
      change (p (C1,C2)). apply (RTC_ind (FS:=D) (FE:=RCList D)); auto.
      intros D' [A B] H K [I|[I|[C [I1 [I2 I3]]]]].
      + unfold RCList in I. apply (RList_in D (A,B)) in I as [_ [_ I]]. auto.
      + subst. unfold p. destruct A. simpl. auto.
      + apply H in I2. apply H in I3.
        destruct A as [Aa Ab], B as [Ba Bb], C as [Ca Cb]. simpl in *.
        intros s J. apply I3, I2, J.
    - intros [E1 [E2 E3]]. apply RTC_closure.
      + apply in_prod_iff. auto.
      + left. apply RList_closure.
        * apply in_prod_iff. auto.
        * simpl. auto.
  Qed.

End PContextRelation.

Record FiniteKripkeModel : Type :=
  mkFiniteKripkeModel {
    WF: Type;
    WS: list WF;
    WE: list (WF * WF);
    eq_decW: eq_dec WF;
    WER := RTC WS WE;
    interp: var -> list WF;
    persist: forall x a b, a el WS -> b el WS
                  -> a el interp x -> (a,b) el WER -> b el interp x
}.

Section FiniteKripke.

  Variable FKM: FiniteKripkeModel.

  Implicit Type x y z: WF FKM.

  Fixpoint evalFK s: list (WF FKM) :=
    match s with
    | Var x => interp FKM x
    | Fal => []
    | And s1 s2 => inter (eq_decX:=@eq_decW FKM) (evalFK s1) (evalFK s2)
    | Or s1 s2 => union (eq_decX:=@eq_decW FKM) (evalFK s1) (evalFK s2)
    | Imp s1 s2 =>
        FKImp (eq_decF:=@eq_decW FKM) (WS FKM) (WER FKM) (evalFK s1) (evalFK s2)
    end.

  Definition satisFK x s := x el evalFK s.
  Definition satisFK' x A := forall s, s el A -> satisFK x s.

  Lemma evalFK_mono x y s:
    x el WS FKM -> y el WS FKM -> satisFK x s -> (x,y) el WER FKM -> satisFK y s.
  Proof.
    intros D1 D2 D E.
    induction s as [v| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
    unfold satisFK in *; simpl in *; auto.
    - apply (persist D1 D2 D E).
    - apply FKImp_in. split; auto.
      intros z F. apply inter_sound in F as [F1 F2].
      apply (UCSE_sound (eq_decF:=@eq_decW FKM) (WS FKM) (WE FKM)) in F1.
      apply FKImp_in in D as [_ D]. apply D.
      apply inter_sound. split; auto.
      apply UCSE_sound. apply (RTC_transitive (y:=y)); auto.
      apply RTC_in in F1 as [_ F1]. exact F1.
    - apply inter_sound in D as [D3 D4]. apply inter_sound. auto.
    - apply union_sound. apply union_sound in D as [D|D]; auto.
  Qed.

  Lemma evalFK_OrAR x B:
    (forall s, s el B -> ~satisFK x s) -> ~satisFK x (OrAR B).
  Proof.
    induction B as [|t B].
    - simpl. auto.
    - intros E. intros F. unfold satisFK in F. destruct B as [|u B].
      + simpl in F. apply (E t); auto.
      + simpl in F. apply union_sound in F as [F|F].
        * apply (E t); auto.
        * apply IHB; auto.
  Qed.

  Lemma nd_soundFK A s:
    nd A s -> forall x, x el WS FKM -> satisFK' x A -> satisFK x s.
  Proof.
    intros C. induction C as [A s D|A s D IH|A s t D IH|A s t D IHD E IHE|
                              A s t D IHD E IHE|A s t u D IHD E IHE|A s t D IH|
                              A s t D IH|A s t u D IHD E IHE F IHF];
    intros x H G; auto.
    - exfalso. apply (IH x H G).
    - unfold satisFK. simpl. apply FKImp_in. split; auto.
      intros y E. apply inter_sound in E as [E1 E2].
      apply (UCSE_sound (eq_decF:=@eq_decW FKM) (WS FKM) (WE FKM)) in E1.
      apply IH.
      + apply RTC_in in E1 as [_ E1]. auto.
      + intros s' [K|K]; subst; auto.
        apply (evalFK_mono H); auto. apply RTC_in in E1 as [_ E1]. auto.
    - specialize (IHD x H G). specialize (IHE x H G). unfold satisFK in *.
      simpl in IHD. apply FKImp_in in IHD as [_ IHD].
      apply IHD. apply inter_sound. split; auto.
      apply UCSE_sound, RTC_reflexive, H.
    - specialize (IHD x H G). specialize (IHE x H G). unfold satisFK. simpl.
      apply inter_sound. auto.
    - specialize (IHD x H G). unfold satisFK in *. simpl in IHD.
      apply inter_sound in IHD as [D1 D2]. apply IHE; auto.
      intros v [K|[K|K]]; subst; auto.
    - specialize (IH x H G). apply union_sound. auto.
    - specialize (IH x H G). apply union_sound. auto.
    - specialize (IHD x H G). unfold satisFK in *. simpl in IHD.
      apply union_sound in IHD as [IHD|IHD].
      + apply (IHE x H). intros v [K|K]; subst; auto.
      + apply (IHF x H). intros v [K|K]; subst; auto.
  Qed.

  Let W: Type := {x | x el WS FKM}.

  Let pW := fun x => x el WS FKM.

  Implicit Type u v w: W.

  Let WR u v: Prop :=
    let (x, Ex) := u in let (y, Ey) := v in (x,y) el WER FKM.

  Let WRref: reflexive _ WR.
  Proof. intros [x Ex]. apply RTC_reflexive, Ex. Qed.

  Let WRtra: transitive _ WR.
  Proof. intros [x Ex] [y Ey] [z Ez]. apply RTC_transitive; auto. Qed.

  Let interp (x: var) (w: W): Prop :=
    let (y, _) := w in y el (@interp FKM) x.

  Definition FK2K : KripkeModel.
  Proof.
    apply (@mkKripkeModel W WR WRref WRtra interp).
    intros x [y Ey] [z Ez]. simpl. apply (@persist FKM); auto.
  Defined.

End FiniteKripke.


Section SubClause.
  Definition trip (S: sform) : form := match S with | +s | -s => s end.

  Instance positive_dec: forall S, dec (positive S).
  Proof. intros [s|s]; auto. Defined.

  Instance negative_dec: forall S, dec (negative S).
  Proof. intros [s|s]; auto. Defined.

  Definition C2PC (C: clause) : PContext :=
    (map trip (filter positive C), map trip (filter negative C)).

  Lemma postrip_in s C:
    +s el C <-> s el map trip (filter positive C).
  Proof. split; intros E.
    - apply in_map_iff. exists (+s). split; auto.
      apply in_filter_iff. simpl. auto.
    - apply in_map_iff in E as [S [E F]]. apply in_filter_iff in F as [F G].
      destruct S; subst; auto.
      exfalso. exact G.
  Qed.

  Lemma negtrip_in s C:
    -s el C <-> s el map trip (filter negative C).
  Proof. split; intros E.
    - apply in_map_iff. exists (-s). split; auto.
      apply in_filter_iff. simpl. auto.
    - apply in_map_iff in E as [S [E F]]. apply in_filter_iff in F as [F G].
      destruct S; simpl in E, G; subst; auto.
      exfalso. exact G.
  Qed.

  Definition subPC (C C': PContext): Prop :=
    let (A,B) := C in let (A',B') := C' in A <<= A' /\ B <<= B'.

  Lemma subC (C C': clause):
    C <<= C' <-> subPC (C2PC C) (C2PC C').
  Proof.
    split.
    - intros E. split; intros S F; apply in_map_iff;
      apply in_map_iff in F as [S' [F G]]; exists S'; split; auto; revert G;
      apply filter_mono, E.
    - intros [E F] S G. destruct S as [s|s].
      + apply postrip_in in G. apply E in G. apply postrip_in in G. exact G.
      + apply negtrip_in in G. apply F in G. apply negtrip_in in G. exact G.
  Qed.


  Definition PC2C (A B: context): clause :=
      map (fun (s: form) => +s) A ++ map (fun (s: form) => -s) B.

  Lemma PC2C_pos A B s:
    s el A <-> +s el PC2C A B.
  Proof. unfold PC2C. split; intros E.
    - apply in_app_iff. left. apply in_map_iff. exists s. auto.
    - apply in_app_iff in E as [E|E];
      apply in_map_iff in E as [s' [E F]]; inv E; auto.
  Qed.

  Lemma PC2C_neg A B s:
    s el B <-> -s el PC2C A B.
  Proof. unfold PC2C. split; intros E.
    - apply in_app_iff. right. apply in_map_iff. exists s. auto.
    - apply in_app_iff in E as [E|E];
      apply in_map_iff in E as [s' [E F]]; inv E; auto.
  Qed.

  Lemma PC2C_pushneg A B s: PC2C A (s :: B) <<= -s :: PC2C A B.
  Proof.
    unfold PC2C. intros S E. apply in_app_iff in E as [E|E];
    apply in_map_iff in E as [s' [E F]].
    - subst. right. apply PC2C_pos, F.
    - destruct F as [F|F]; subst; auto. right. apply PC2C_neg, F.
  Qed.

  Lemma PC2C_pushpos A B s: PC2C (s :: A) B <<= +s :: PC2C A B.
  Proof.
    unfold PC2C. intros S E. apply in_app_iff in E as [E|E];
    apply in_map_iff in E as [s' [E F]].
    - destruct F as [F|F]; subst; auto. right. apply PC2C_pos, F.
    - subst. right. apply PC2C_neg, F.
  Qed.

  Definition pos (C: clause) : clause := filter positive C.

  Lemma negtrip_empty C: map trip (filter negative (pos C)) = [].
  Proof.
    apply incl_nil_eq. intros S E.
    apply in_map_iff in E as [[s|s] [E F]];
    apply in_filter_iff in F as [F G]; auto.
    simpl. clear E G S. induction C as [|[s'|s'] C]; auto.
    destruct F as [F|F]; subst; auto; discriminate F.
  Qed.

  (* D supports C if some C' el D and tab C', since tab C' -> tab C by tabW *)
  Definition sup (D: list clause) (C: clause) : Prop :=
      exists C', C' el D /\ C' <<= C.

  Definition stepTab' (D: list clause) (S: sform) (C: clause): Prop :=
    match S with
    | +Var x => -Var x el C
    | +Fal   => True
    | +Imp s t => sup D (-s::C) /\ sup D (+t :: C)
    | -Imp s t => sup D (+s::-t::pos C)
    | +And s t => sup D (+s::+t::C)
    | -And s t => sup D (-s::C) /\ sup D (-t::C)
    | +Or  s t => sup D (+s::C) /\ sup D (+t::C)
    | -Or  s t => sup D (-s::-t::C)
    | _ => False
    end.

  Definition stepTab (D: list clause) (C: clause) : Prop :=
    exists S, S el C /\ stepTab' D S C.

  Instance stepTab_dec D C: dec (stepTab D C).
  Proof.
    unfold stepTab. apply list_exists_dec. intros [[| | | |]|[| | | |]];
    unfold stepTab', sup; auto.
  Defined.

  Lemma stepTab_mono D D' C:
    D <<= D' -> stepTab D C -> stepTab D' C.
  Proof.
    intros E [S [F G]]. exists S. split; auto. unfold stepTab', sup in *.
    destruct S as [[x| |s t|s t|s t]|[x| |s t|s t|s t]]; auto.
    - destruct G as [[C1 [G1 G2]] [C2 [G3 G4]]]. split.
      + exists C1. auto.
      + exists C2. auto.
    - destruct G as [C' [G1 G2]]. exists C'. auto.
    - destruct G as [[C1 [G1 G2]] [C2 [G3 G4]]]. split.
      + exists C1. auto.
      + exists C2. auto.
    - destruct G as [C' [G1 G2]]. exists C'. auto.
    - destruct G as [[C1 [G1 G2]] [C2 [G3 G4]]]. split.
      + exists C1. auto.
      + exists C2. auto.
    - destruct G as [C' [G1 G2]]. exists C'. auto.
  Qed.

End SubClause.

Section TableauxDecidability.

  Variable U: clause.
  Variable sscU: ss_closed U.

  Definition lfpTab :=
    FCI.C (step_dec:=stepTab_dec) (step:=stepTab) (power U).

  Lemma lfpTab_closure (C: clause):
    C el power U -> stepTab lfpTab C -> C el lfpTab.
  Proof. apply FCI.closure. Qed.

  Lemma lfpTab_ind (p: clause -> Prop):
    (forall D C, inclp D p -> C el power U -> stepTab D C -> p C)
    -> inclp lfpTab p.
  Proof. apply FCI.ind. Qed.

  Lemma lfp_tab (C: clause):
    C el lfpTab -> let (A, B) := (C2PC C) in tab A B.
  Proof.
    intros E. unfold C2PC. pattern C. apply lfpTab_ind; auto.
    intros D C' F G H. unfold stepTab, stepTab' in H.
    destruct H as [[[x| |s t|s t|s t]|[x| |s t|s t|s t]] [H1 H2]].
    - apply (tabC (x:=x)).
      + apply postrip_in, H1.
      + apply negtrip_in, H2.
    - apply tabF. apply postrip_in, H1.
    - destruct H2 as [[C1 [K1 K2]] [C2 [K3 K4]]].
      apply (tabIP (s:=s) (t:=t)).
      + apply postrip_in, H1.
      + apply F in K1. apply subC in K2 as [K2 K5]. apply (tabW K1); auto.
      + apply F in K3. apply subC in K4 as [K4 K5]. apply (tabW K3); auto.
    - destruct H2 as [C1 [H2 H3]].
      apply F in H2. apply subC in H3 as [H3 H4]. apply (tabAP (s:=s) (t:=t)).
      + apply postrip_in, H1.
      + apply (tabW H2); auto.
    - destruct H2 as [[C1 [K1 K2]] [C2 [K3 K4]]].
      apply (tabOP (s:=s) (t:=t)).
      + apply postrip_in, H1.
      + apply F in K1. apply subC in K2 as [K2 K5]. apply (tabW K1); auto.
      + apply F in K3. apply subC in K4 as [K4 K5]. apply (tabW K3); auto.
    - exfalso. apply H2.
    - exfalso. apply H2.
    - destruct H2 as [C1 [H2 H3]].
      apply F in H2. apply subC in H3 as [H3 H4]. simpl in H3, H4.
      rewrite negtrip_empty in H4. unfold pos in H3.
      rewrite filter_idempotent in H3. apply (tabIN (s:=s) (t:=t)).
      + apply negtrip_in, H1.
      + apply (tabW H2); auto.
    - destruct H2 as [[C1 [K1 K2]] [C2 [K3 K4]]].
      apply (tabAN (s:=s) (t:=t)).
      + apply negtrip_in, H1.
      + apply F in K1. apply subC in K2 as [K2 K5]. apply (tabW K1); auto.
      + apply F in K3. apply subC in K4 as [K4 K5]. apply (tabW K3); auto.
    - destruct H2 as [C1 [H2 H3]].
      apply F in H2. apply subC in H3 as [H3 H4].
      apply (tabON (s:=s) (t:=t)).
      + apply negtrip_in, H1.
      + apply (tabW H2); auto.
  Qed.

  Lemma tab_lfp A B:
    PC2C A B <<= U -> tab A B -> rep (PC2C A B) U el lfpTab.
  Proof.
    intros E F.
    induction F as [A B F|A B x F G|A B s t F _ IH1 _ IH2|
                    A B s t F _ IH|A B s t F _ IH|
                    A B s t F _ IH1 _ IH2|A B s t F _ IH1 _ IH2|
                    A B s t F _ IH]; unfold stepTab;
    assert (H: rep (PC2C A B) U === PC2C A B) by (apply rep_equi, E);
    apply lfpTab_closure; try apply rep_power.
    - exists (+Fal). simpl. split; auto. rewrite H. apply PC2C_pos, F.
    - exists (+Var x). simpl. rewrite H. split.
      + apply PC2C_pos, F.
      + apply PC2C_neg, G.
    - assert (K: -s el U /\ +t el U).
        { apply PC2C_pos with (B:=B) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C A (s :: B) <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply E. apply in_app_iff. left. exact G.
          - apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_neg, K. }
      assert (K2: PC2C (t::A) B <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          + apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_pos, K.
          + apply E. apply in_app_iff. right. exact G. }
      exists (+Imp s t). simpl. unfold sup. rewrite H. repeat split.
      + apply PC2C_pos, F.
      + exists (rep (PC2C A (s :: B)) U). split; auto.
        rewrite (rep_equi K1), (rep_equi E). apply PC2C_pushneg.
      + exists (rep (PC2C (t :: A) B) U). split; auto.
        rewrite (rep_equi K2), (rep_equi E). apply PC2C_pushpos.
    - assert (K: +s el U /\ -t el U).
        { apply PC2C_neg with (A:=A) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C (s :: A) [t] <<= U).
        { intros S G. apply in_app_iff in G as [G|G];
          apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
          - apply E, PC2C_pos, K.
          - exfalso. apply K. }
      exists (-Imp s t). simpl. unfold sup. split.
      + rewrite H. apply PC2C_neg, F.
      + exists (rep (PC2C (s :: A) [t]) U). split; auto.
        rewrite (rep_equi K1). intros S G.
        apply in_app_iff in G as [G|G]; 
        apply in_map_iff in G as [s' [G [I|I]]]; subst; auto.
        * right. right. apply in_filter_iff. simpl. split; auto.
          rewrite H. apply PC2C_pos, I.
        * exfalso. apply I.
    - assert (K: +s el U /\ +t el U).
        { apply PC2C_pos with (B:=B) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C (s :: t :: A) B <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply in_map_iff in G as [s' [G [K|[K|K]]]]; subst; auto.
            apply E, PC2C_pos, K.
          - apply E. apply in_app_iff. right. exact G. }
      exists (+And s t). simpl. split.
      + rewrite H. apply PC2C_pos, F.
      + exists (rep (PC2C (s :: t :: A) B) U). split; auto.
        rewrite (rep_equi K1), (rep_equi E). apply PC2C_pushpos.
    - assert (K: -s el U /\ -t el U).
        { apply PC2C_neg with (A:=A) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C A (s :: B) <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply E, in_app_iff. auto.
          - apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_neg, K. }
      assert (K2: PC2C A (t :: B) <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply E, in_app_iff. auto.
          - apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_neg, K. }
      exists (-And s t). simpl. repeat split.
      + rewrite H. apply PC2C_neg, F.
      + exists (rep (PC2C A (s :: B)) U). split; auto.
        rewrite (rep_equi K1), (rep_equi E). apply PC2C_pushneg.
      + exists (rep (PC2C A (t :: B)) U). split; auto.
        rewrite (rep_equi K2), (rep_equi E). apply PC2C_pushneg.
    - assert (K: +s el U /\ +t el U).
        { apply PC2C_pos with (B:=B) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C (s :: A) B <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_pos, K.
          - apply E, in_app_iff. auto. }
      assert (K2: PC2C (t :: A) B <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply in_map_iff in G as [s' [G [K|K]]]; subst; auto.
            apply E, PC2C_pos, K.
          - apply E, in_app_iff. auto. }
      exists (+Or s t). simpl. repeat split.
      + rewrite H. apply PC2C_pos, F.
      + exists (rep (PC2C (s::A) B) U). split; auto.
        rewrite (rep_equi K1), (rep_equi E). apply PC2C_pushpos.
      + exists (rep (PC2C (t::A) B) U). split; auto.
        rewrite (rep_equi K2), (rep_equi E). apply PC2C_pushpos.
    - assert (K: -s el U /\ -t el U).
        { apply PC2C_neg with (A:=A) in F. apply E in F.
          apply sscU in F. exact F. } destruct K as [H1 H2].
      assert (K1: PC2C A (s :: t :: B) <<= U).
        { intros S G. apply in_app_iff in G as [G|G].
          - apply E. apply in_app_iff. auto.
          - apply in_map_iff in G as [s' [G [K|[K|K]]]]; subst; auto.
            apply E, PC2C_neg, K. }
      exists (-Or s t). simpl. split.
      + rewrite H. apply PC2C_neg, F.
      + exists (rep (PC2C A (s :: t :: B)) U). split; auto.
        rewrite (rep_equi K1), (rep_equi E). rewrite PC2C_pushneg.
        apply incl_shift. apply PC2C_pushneg.
  Qed.

  Lemma tab_dec' A B:
    PC2C A B <<= U -> dec (tab A B).
  Proof.
    intros E. decide (rep (PC2C A B) U el lfpTab) as [H|H].
    - left. apply lfp_tab in H. unfold C2PC in H.
      apply (tabW H); intros s F; apply in_map_iff in F as [S [F G]];
      apply in_filter_iff in G as [G K]; rewrite (rep_equi E) in G;
      destruct S as [s'|s']; simpl in F; subst.
      + apply PC2C_pos in G. exact G.
      + exfalso. apply K.
      + exfalso. apply K.
      + apply PC2C_neg in G. exact G.
    - right. intros F. apply tab_lfp in F; auto.
  Defined.

End TableauxDecidability.

Section TableauxConsistency.

  Theorem tab_dec A B: dec (tab A B).
  Proof.
    destruct (ssL_correct (PC2C A B)) as [E F].
    apply (tab_dec' E F).
  Defined.

  Definition cons (C: PContext) :=
    let (A,B) := C in ~tab A B.

  Instance cons_dec: forall C, dec (cons C).
  Proof.
    intros [A B]. unfold cons. apply impl_dec; auto. apply tab_dec.
  Defined.

  Lemma consW: forall C C', cons C -> subPC C' C -> cons C'.
  Proof.
    intros [A B] [A' B']. unfold cons, subPC.
    intros E [F G] H. apply E. apply (tabW H); auto.
  Qed.

End TableauxConsistency.
