Require Export PropiL.
Require Export Coq.Relations.Relation_Definitions.

(** Heyting Algebras as semantics for Intuitionistic Logic *)

Record HeytingAlgebra : Type :=
  mkHeytingAlgebra {
    H:Type;
    R: H -> H -> Prop;
    bot: H;
    meet: H -> H -> H;
    join: H -> H -> H;
    imp: H -> H -> H;
    Rref : reflexive _ R;
    Rtra : transitive _ R;
    bot1 : forall u, R bot u;
    meet1 : forall u v w, R w u /\ R w v <-> R w (meet u v);
    join1 : forall u v w, R u w /\ R v w <-> R (join u v) w;
    imp1 : forall u v w, R (meet w u) v <-> R w (imp u v)
}.

Section HAProperty.

  Variable HA: HeytingAlgebra.
  Implicit Type u v w: H HA.

  Definition eqH u v := R u v /\ R v u.

  Lemma meet2 u v: R (meet u v) u.
  Proof.
    assert (H:= meet1 u v (meet u v)).
    assert (K:= Rref (meet u v)).
    apply H in K as [K _]. apply K.
  Qed.

  Lemma meet3 u v: R (meet u v) v.
  Proof.
    assert (H:= meet1 u v (meet u v)).
    assert (K:= Rref (meet u v)).
    apply H in K as [_ K]. apply K.
  Qed.

  Lemma meet_com u v: eqH (meet u v) (meet v u).
  Proof. split; apply meet1; split; auto using meet2, meet3. Qed.

  Definition top := imp (bot HA) (bot HA).

  Lemma top1 u: R u top.
  Proof. apply imp1, meet3. Qed.

  Lemma join2 u v: R u (join u v).
  Proof.
    assert (H:= join1 u v (join u v)).
    assert (K:= Rref (join u v)).
    apply H in K as [K _]. apply K.
  Qed.

  Lemma join3 u v: R v (join u v).
  Proof.
    assert (H:= join1 u v (join u v)).
    assert (K:= Rref (join u v)).
    apply H in K as [_ K]. apply K.
  Qed.

  Lemma join_com u v: eqH (join u v) (join v u).
  Proof. split; apply join1; split; auto using join2, join3. Qed.

  Lemma imp2 u v: R (meet (imp u v) u) v.
  Proof. apply imp1, Rref. Qed.

End HAProperty.

(** * Soundness of Heyting algebra *)

Section HASound.

  Variable HA : HeytingAlgebra.

  Variable interp : var -> H HA.

  Fixpoint evalH (s : form) : H HA :=
    match s with
      | Var x => interp x
      | Fal => bot HA
      | Imp s t => imp (evalH s) (evalH t)
      | And s t => meet (evalH s) (evalH t)
      | Or s t => join (evalH s) (evalH t)
    end.

  Fixpoint evalH' (A : list form) : H HA :=
    match A with
      | nil => top HA
      | s::A => meet (evalH' A) (evalH s)
    end.
 
  Definition hent A s := R (evalH' A) (evalH s).

  Lemma nd_soundHA A s : nd A s -> hent A s.
  Proof.
    unfold hent.
    intros C. induction C as [A s D|A s D IH|A s t D IH|A s t D IHD E IHE|
                              A s t D IHD E IHE|A s t u D IHD E IHE|A s t D IH|
                              A s t D IH|A s t u D IHD E IHE F IHF].
    - induction A as [|t A].
      + exfalso. exact D.
      + destruct D as [D|D].
        * subst t. simpl. apply meet3.
        * simpl. apply Rtra with (y := (evalH' A)); auto using meet2.
    - apply (Rtra IH), bot1.
    - simpl. apply imp1. exact IH.
    - apply Rtra with (y := (meet (imp (evalH s) (evalH t)) (evalH s))).
      + apply meet1. auto.
      + apply imp2.
    - simpl. apply meet1. auto.
    - apply Rtra with (y := evalH' (s :: t :: A)); auto.
      apply Rtra with (y := meet (evalH' A) (evalH (And s t))).
      + apply meet1. split. apply Rref. apply IHD.
      + simpl. apply meet1. split.
        * apply meet1. split.
          { apply meet2. } 
          { apply Rtra with (y:= meet (evalH s) (evalH t)); apply meet3. }
        * apply Rtra with (y:= meet (evalH s) (evalH t));
          auto using meet2, meet3.
    - apply (Rtra IH); auto. apply join2.
    - apply (Rtra IH); auto. apply join3.
    - apply Rtra with (y:= meet (evalH' A) (evalH (Or s t))).
      + apply meet1. split. apply Rref. apply IHD.
      + apply Rtra with (y:= join (evalH' (s :: A)) (evalH' (t :: A))).
        * apply Rtra with
          (y:= meet (imp (evalH' A) (join (evalH' (s :: A)) (evalH' (t :: A))))
                    (evalH' A)); auto using imp2.
          apply meet1; split; auto using meet2.
          apply Rtra with (y:= evalH (Or s t)); auto using meet3.
          simpl. apply join1; split; apply imp1.
          { apply Rtra with (y:=meet (evalH' A) (evalH s)); auto using join2.
            apply meet1; auto using meet2, meet3. }
          { apply Rtra with (y:=meet (evalH' A) (evalH t)); auto using join3.
            apply meet1; auto using meet2, meet3. }
        * apply join1. auto.
  Qed.

  Lemma HAequiv s t : nd [] (Equi s t) -> eqH (evalH s) (evalH t).
  Proof.
    intros C. apply nd_soundHA in C. simpl in C. split.
    - apply (@Rtra HA (evalH s) (meet (imp (evalH s) (evalH t)) (evalH s))).
      + apply meet1. split.
        * apply (Rtra (top1 (evalH s))). apply (Rtra C). apply meet2.
        * apply Rref.
      + apply imp2.
    - apply (@Rtra HA (evalH t) (meet (imp (evalH t) (evalH s)) (evalH t))).
      + apply meet1. split.
        * apply (Rtra (top1 (evalH t))). apply (Rtra C). apply meet3.
        * apply Rref.
      + apply imp2.
  Qed.

End HASound.

(** * Completeness of Heyting algebra *)

Section FormHeytingAlgebra.

  Let H := form.

  Let R s t := nd [] (Imp s t).

  Let bot := Fal.

  Let meet s t := And s t.

  Let join s t := Or s t.

  Let imp s t := Imp s t.

  Definition FormHA: HeytingAlgebra.
    apply (@mkHeytingAlgebra H R bot meet join imp); unfold R.
    - intros w. auto.
    - intros u v w C D. apply ndII. apply nd_cut with (s:=v).
      + apply ndIE with (s:=u) (t:=v); auto.
        apply (ndW C); auto.
      + apply ndW with (A:=[v]); auto. apply ndIE with (s:=v) (t:=w); auto.
        apply (ndW D); auto.
    - intros u. apply ndII, ndE. auto.
    - intros u v w. unfold meet. split.
      + intros [C D]. apply ndII, ndAI.
        * apply ndIE with (s:=w) (t:=u); auto. apply (ndW C); auto.
        * apply ndIE with (s:=w) (t:=v); auto. apply (ndW D); auto.
      + intros C. split;
        apply ndII; apply ndAE with (s:=u) (t:=v); auto;
        apply ndIE with (s:=w) (t:=And u v); auto;
        apply (ndW C); auto.
    - intros u v w. split.
      + intros [C D]. apply ndII. apply ndOE with (s:=u) (t:=v); auto.
        * apply ndW with (A:=[u]); auto. apply ndIE with (s:=u) (t:=w); auto.
          apply (ndW C); auto.
        * apply ndW with (A:=[v]); auto. apply ndIE with (s:=v) (t:=w); auto.
          apply (ndW D); auto.
      + intros C. split; apply ndII; apply nd_cut with (s:=Or u v); auto;
        apply ndIE with (s:=Or u v) (t:=w); auto; apply (ndW C); auto.
    - intros u v w. split; intros C.
      + apply ndII, ndII. apply nd_cut with (s:=And w u).
        * apply ndAI; auto.
        * apply ndIE with (s:=And w u); auto. apply (ndW C); auto.
      + apply ndII. apply ndAE with (s:=w) (t:=u); auto.
        apply ndIE with (s:=u) (t:=v); auto.
        apply ndIE with (s:=w) (t:=Imp u v); auto.
        apply (ndW C); auto.
  Defined.

  Definition Vf (x: var) := (Var x).

  Lemma evalHf s: (@evalH FormHA Vf s) = s.
  Proof.
    induction s as [x| |s IHs t IHt|s IHs t IHt|s IHs t IHt]; auto; simpl;
    rewrite IHs, IHt; reflexivity.
  Qed.

End FormHeytingAlgebra.

Lemma HA_nd s:
  (forall (HA: HeytingAlgebra) (V: var -> H HA), R (top HA) (evalH V s))
  -> nd [] s.
Proof.
  intros C.
  specialize (C FormHA Vf). rewrite evalHf in C. simpl in C.
  apply nd_cut with (s:=Imp Fal Fal); auto.
  apply ndIE with (s:=Imp Fal Fal); auto.
  apply (ndW C); auto.
Qed.

Fixpoint ltr (A: context) (s: form): form :=
  match A with 
  | [] => s
  | t::A => ltr A (Imp t s)
  end.

Lemma ltr_CharImp E A s:
  CharImp E -> (E A s <-> E [] (ltr A s)).
Proof.
  intros CImp. revert s. induction A as [|t A]; intros s.
  - reflexivity.
  - simpl. split; intros C.
    + apply CImp in C. apply IHA, C.
    + apply IHA in C. apply CImp, C.
Qed.

Lemma hent_imp (HA: HeytingAlgebra) V: CharImp (@hent HA V).
Proof. split; unfold hent; simpl; intros C; apply imp1, C. Qed.

(** Completeness of Heyting Algebras *)
Theorem HA_iff_nd A s:
  (forall (HA: HeytingAlgebra) (V: var -> H HA), hent V A s)
  <-> nd A s.
Proof. split.
  - intros C. apply (ltr_CharImp A s nd_imp).
    apply HA_nd. intros HA V. apply (ltr_CharImp A s (hent_imp V)).
    exact (C HA V).
  - intros C HA V. apply nd_soundHA, C.
Qed.


(** * Kripke Semantics *)

Record KripkeModel : Type :=
  mkKripkeModel {
    W: Type;
    WR: W -> W -> Prop;
    WRref : reflexive _ WR;
    WRtra : transitive _ WR;
    pred := W -> Prop;
    interp: var -> pred;
    persist: forall x a b, interp x a -> WR a b -> interp x b
}.

Section KripkeSemantics.
  Variable KM: KripkeModel.

  Fixpoint evalK s: pred KM :=
    match s with
    | Var x => interp x
    | Fal => fun _ => False
    | And s1 s2 => fun w => evalK s1 w /\ evalK s2 w
    | Or s1 s2 => fun w => evalK s1 w \/ evalK s2 w
    | Imp s1 s2 => fun w => forall v, WR w v -> evalK s1 v -> evalK s2 v
    end.

  Lemma evalK_mono u v s: evalK s u -> WR u v -> evalK s v.
  Proof.
    intros D E. 
    induction s as [x| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
    simpl in *; auto.
    - apply (persist D E).
    - intros w F G. apply D; auto. apply (WRtra E F).
    - destruct D as [D1 D2]; auto.
    - destruct D as [D|D]; auto.
  Qed.

  Definition evalKA A w := forall s, s el A -> evalK s w.

  Lemma nd_soundKM A s: nd A s -> forall w, evalKA A w -> evalK s w.
  Proof.
    intros C. induction C as [A s D|A s D IH|A s t D IH|A s t D IHD E IHE|
                              A s t D IHD E IHE|A s t u D IHD E IHE|A s t D IH|
                              A s t D IH|A s t u D IHD E IHE F IHF];
    intros w G; auto.
    - exfalso. apply (IH w G).
    - intros u E F. apply IH.
      intros s' [H|H]; subst; auto. apply evalK_mono with (u:=w); auto.
    - specialize (IHD w G). specialize (IHE w G). apply IHD.
      apply WRref. apply IHE.
    - specialize (IHD w G). specialize (IHE w G). simpl. auto.
    - destruct (IHD w G) as [D1 D2]. apply IHE.
      intros v [H|[H|H]]; subst; auto.
    - specialize (IH w G). simpl. auto.
    - specialize (IH w G). simpl. auto.
    - destruct (IHD w G) as [H|H].
      + apply IHE. intros v [K|K]; subst; auto.
      + apply IHF. intros v [K|K]; subst; auto.
  Qed.

End KripkeSemantics.

(** * From Kripke to Heyting *)
Section Kripke2Heyting.

  Variable KM : KripkeModel.

  Let UCS: pred KM -> Prop
    := fun p => forall u v: W KM, WR u v -> p u -> p v.

  Let H: Type := {p: pred KM | UCS p}.

  Let R: H -> H -> Prop 
    := fun a b => let (p,_) := a in let (q,_) := b in forall u, p u -> q u.

  Let bot: H.
    exists (fun w => False). intros u v. intros _ F. exact F.
  Defined.

  Let meet: H -> H -> H.
    intros [p Dp] [q Dq].
    exists (fun w => p w /\ q w).
    intros u v C [Ep Eq]. split.
    - apply (Dp u v). exact C. exact Ep.
    - apply (Dq u v). exact C. exact Eq.
  Defined.

  Let join: H -> H -> H.
    intros [p Dp] [q Dq].
    exists (fun w => p w \/ q w).
    intros u v C [E|E].
    - left. apply (Dp u v). exact C. exact E.
    - right. apply (Dq u v). exact C. exact E.
  Defined.

  Let UC (w: W KM) : W KM -> Prop := fun v => WR w v.

  (* imp s t := {w | (UCS of w) inter s is subset of t } *)
  Let imp: H -> H -> H.
    intros [p Dp] [q Dq].
    exists (fun w => forall w', UC w w' /\ p w' -> q w').
    intros u v C E w' [F G]. apply E. split.
    - apply (WRtra C F).
    - exact G.
  Defined.

  Definition K2H: HeytingAlgebra.
  apply (@mkHeytingAlgebra H R bot meet join imp).
  - intros [p Dp]. simpl. auto.
  - intros [p Dp] [q Dq] [r Dr]. simpl. auto.
  - intros [p Dp]. simpl. tauto.
  - intros [p Dp] [q Dq] [r Dr]. simpl. split.
    + intros [E F]. auto.
    + intros E. split; intros u F; destruct (E u F); auto.
  - intros [p Dp] [q Dq] [r Dr]. simpl. split.
    + intros [E F] u [G|G]; auto.
    + intros E. split; intros u F; auto.
  - intros [p Dp] [q Dq] [r Dr]. simpl. split.
    + intros E u F w' [G I]. apply E. split; auto. apply (Dr u w' G F).
    + intros E u [F G]. apply (E u F); split; unfold UC; auto; apply WRref.
  Defined.

  Definition K2Hinterp: var -> H.
    intros x. exists (interp x).
    intros u v E F. apply (persist F E).
  Defined.

  (* evalK s = evalH s *)
  Theorem evalK_evalH_agree s:
    let (p, _) := @evalH K2H K2Hinterp s in
    forall w, evalK s w <-> p w.
  Proof.
    induction s as [x| |s IHs t IHt|s IHs t IHt|s IHs t IHt]; simpl; try tauto;
    destruct (@evalH K2H K2Hinterp s) as [p Dp];
    destruct (@evalH K2H K2Hinterp t) as [q Dq]; simpl.
    - intros w. split; intros C.
      + intros w' [E F]. apply IHt. apply (C w' E). apply IHs, F.
      + intros v E F. apply IHt. apply C. split; auto. apply IHs, F.
    - intros w. split; intros [E F];
      split; try (apply IHs, E); try (apply IHt, F).
    - intros w. split; intros [E|E];
      try (left; apply IHs, E); try (right; apply IHt, E).
  Qed.

End Kripke2Heyting.
